# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import inspect
import logging

import requests
import xmltodict
from xml.parsers.expat import ExpatError
from optionaldict import optionaldict

from wechatpyyyy.crypto import WeChatRefundCrypto
from wechatpyyyy.utils import random_string
from wechatpyyyy.exceptions import WeChatPayException, InvalidSignatureException
from wechatpyyyy.pay.utils import (
    calculate_signature, calculate_signature_hmac, _check_signature, dict_to_xml
)
from wechatpyyyy.pay.base import BaseWeChatPayAPI
from wechatpyyyy.pay import api

logger = logging.getLogger(__name__)


def _is_api_endpoint(obj):
    return isinstance(obj, BaseWeChatPayAPI)


class WeChatPay(object):
    """
    微信支付接口

    :param appid: 微信公众号 appid
    :param sub_appid: 当前调起支付的小程序APPID
    :param api_key: 商户 key,不要在这里使用小程序的密钥
    :param mch_id: 商户号
    :param sub_mch_id: 可选，子商户号，受理模式下必填
    :param mch_cert: 必填，商户证书路径
    :param mch_key: 必填，商户证书私钥路径
    :param timeout: 可选，请求超时时间，单位秒，默认无超时设置
    :param sandbox: 可选，是否使用测试环境，默认为 False
    """

    redpack = api.WeChatRedpack()
    """红包接口"""
    transfer = api.WeChatTransfer()
    """企业付款接口"""
    coupon = api.WeChatCoupon()
    """代金券接口"""
    order = api.WeChatOrder()
    """订单接口"""
    refund = api.WeChatRefund()
    """退款接口"""
    micropay = api.WeChatMicroPay()
    """刷卡支付接口"""
    tools = api.WeChatTools()
    """工具类接口"""
    jsapi = api.WeChatJSAPI()
    """公众号网页 JS 支付接口"""
    withhold = api.WeChatWithhold()
    """代扣接口"""

    API_BASE_URL = 'https://api.mch.weixin.qq.com/'

    def __new__(cls, *args, **kwargs):
        self = super(WeChatPay, cls).__new__(cls)
        api_endpoints = inspect.getmembers(self, _is_api_endpoint)
        for name, _api in api_endpoints:
            api_cls = type(_api)
            _api = api_cls(self)
            setattr(self, name, _api)
        return self

    def __init__(self, appid, api_key, mch_id, sub_mch_id=None,
                 mch_cert=None, mch_key=None, timeout=None, sandbox=False, sub_appid=None):
        self.appid = appid
        self.sub_appid = sub_appid
        self.api_key = api_key
        self.mch_id = mch_id
        self.sub_mch_id = sub_mch_id
        self.mch_cert = mch_cert
        self.mch_key = mch_key
        self.timeout = timeout
        self.sandbox = sandbox
        self._sandbox_api_key = None
        self._http = requests.Session()

    def _fetch_sandbox_api_key(self):
        nonce_str = random_string(32)
        sign = calculate_signature({'mch_id': self.mch_id, 'nonce_str': nonce_str}, self.api_key)
        payload = dict_to_xml({
            'mch_id': self.mch_id,
            'nonce_str': nonce_str,
        }, sign=sign)
        headers = {'Content-Type': 'text/xml'}
        api_url = '{base}sandboxnew/pay/getsignkey'.format(base=self.API_BASE_URL)
        response = self._http.post(api_url, data=payload, headers=headers)
        return xmltodict.parse(response.text)['xml'].get('sandbox_signkey')

    def _request(self, method, url_or_endpoint, **kwargs):
        if not url_or_endpoint.startswith(('http://', 'https://')):
            api_base_url = kwargs.pop('api_base_url', self.API_BASE_URL)
            if self.sandbox:
                api_base_url = '{url}sandboxnew/'.format(url=api_base_url)
            url = '{base}{endpoint}'.format(
                base=api_base_url,
                endpoint=url_or_endpoint
            )
        else:
            url = url_or_endpoint

        if isinstance(kwargs.get('data', ''), dict):
            data = kwargs['data']
            if 'mchid' not in data:
                # Fuck Tencent
                data.setdefault('mch_id', self.mch_id)
            data.setdefault('sub_mch_id', self.sub_mch_id)
            data.setdefault('nonce_str', random_string(32))
            data = optionaldict(data)

            if data.get('sign_type', 'MD5') == 'HMAC-SHA256':
                sign = calculate_signature_hmac(data, self.sandbox_api_key if self.sandbox else self.api_key)
            else:
                sign = calculate_signature(data, self.sandbox_api_key if self.sandbox else self.api_key)
            body = dict_to_xml(data, sign)
            body = body.encode('utf-8')
            kwargs['data'] = body

        # 商户证书
        if self.mch_cert and self.mch_key:
            kwargs['cert'] = (self.mch_cert, self.mch_key)

        kwargs['timeout'] = kwargs.get('timeout', self.timeout)
        logger.debug('Request to WeChat API: %s %s\n%s', method, url, kwargs)
        res = self._http.request(
            method=method,
            url=url,
            **kwargs
        )
        try:
            res.raise_for_status()
        except requests.RequestException as reqe:
            raise WeChatPayException(
                return_code=None,
                client=self,
                request=reqe.request,
                response=reqe.response
            )

        return self._handle_result(res)

    def _handle_result(self, res):
        res.encoding = 'utf-8'
        xml = res.text
        logger.debug('Response from WeChat API \n %s', xml)
        try:
            data = xmltodict.parse(xml)['xml']
        except (xmltodict.ParsingInterrupted, ExpatError):
            # 解析 XML 失败
            logger.debug('WeChat payment result xml parsing error', exc_info=True)
            return xml

        return_code = data['return_code']
        return_msg = data.get('return_msg')
        result_code = data.get('result_code')
        errcode = data.get('err_code')
        errmsg = data.get('err_code_des')
        if return_code != 'SUCCESS' or result_code != 'SUCCESS':
            # 返回状态码不为成功
            raise WeChatPayException(
                return_code,
                result_code,
                return_msg,
                errcode,
                errmsg,
                client=self,
                request=res.request,
                response=res
            )
        return data

    def get(self, url, **kwargs):
        return self._request(
            method='get',
            url_or_endpoint=url,
            **kwargs
        )

    def post(self, url, **kwargs):
        return self._request(
            method='post',
            url_or_endpoint=url,
            **kwargs
        )

    def check_signature(self, params):
        return _check_signature(params, self.api_key if not self.sandbox else self.sandbox_api_key)

    @classmethod
    def get_payment_data(cls, xml):
        """
        解析微信支付结果通知，获得appid, mch_id, out_trade_no, transaction_id
        如果你需要进一步判断，请先用appid, mch_id来生成WeChatPay,
        然后用`wechatpay.parse_payment_result(xml)`来校验支付结果

        使用示例::

            from wechatpyyyy.pay import WeChatPay
            # 假设你已经获取了微信服务器推送的请求中的xml数据并存入xml变量
            data = WeChatPay.get_payment_appid(xml)
            {
                "appid": "公众号或者小程序的id",
                "mch_id": "商户id",
            }

        """
        try:
            data = xmltodict.parse(xml)
        except (xmltodict.ParsingInterrupted, ExpatError):
            raise ValueError("invalid xml")
        if not data or 'xml' not in data:
            raise ValueError("invalid xml")
        data = data["xml"]
        return {
            "appid": data["appid"],
            "mch_id": data["mch_id"],
            "out_trade_no": data["out_trade_no"],
            "transaction_id": data["transaction_id"],
        }

    def parse_payment_result(self, xml):
        """解析微信支付结果通知"""
        try:
            data = xmltodict.parse(xml)
        except (xmltodict.ParsingInterrupted, ExpatError):
            raise InvalidSignatureException()

        if not data or 'xml' not in data:
            raise InvalidSignatureException()

        data = data['xml']
        sign = data.pop('sign', None)
        real_sign = calculate_signature(data, self.api_key if not self.sandbox else self.sandbox_api_key)
        if sign != real_sign:
            raise InvalidSignatureException()

        for key in ('total_fee', 'settlement_total_fee', 'cash_fee', 'coupon_fee', 'coupon_count'):
            if key in data:
                data[key] = int(data[key])
        data['sign'] = sign
        return data

    def parse_refund_notify_result(self, xml):
        """解析微信退款结果通知"""
        refund_crypto = WeChatRefundCrypto(self.api_key if not self.sandbox else self.sandbox_api_key)
        data = refund_crypto.decrypt_message(xml, self.appid, self.mch_id)
        for key in ('total_fee', 'settlement_total_fee', 'refund_fee', 'settlement_refund_fee'):
            if key in data:
                data[key] = int(data[key])
        return data

    @property
    def sandbox_api_key(self):
        if self.sandbox and self._sandbox_api_key is None:
            self._sandbox_api_key = self._fetch_sandbox_api_key()

        return self._sandbox_api_key



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from wechatpyyyy.pay.base import BaseWeChatPayAPI


class WeChatRefund(BaseWeChatPayAPI):

    def apply(self, total_fee, refund_fee, out_refund_no, transaction_id=None,
              out_trade_no=None, fee_type='CNY', op_user_id=None,
              device_info=None, refund_account='REFUND_SOURCE_UNSETTLED_FUNDS',
              refund_desc=None, notify_url=None):
        """
        申请退款

        :param total_fee: 订单总金额，单位为分
        :param refund_fee: 退款总金额，单位为分
        :param out_refund_no: 商户系统内部的退款单号，商户系统内部唯一，同一退款单号多次请求只退一笔
        :param transaction_id: 可选，微信订单号
        :param out_trade_no: 可选，商户系统内部的订单号，与 transaction_id 二选一
        :param fee_type: 可选，货币类型，符合ISO 4217标准的三位字母代码，默认人民币：CNY
        :param op_user_id: 可选，操作员帐号, 默认为商户号
        :param device_info: 可选，终端设备号
        :param refund_desc: 可选，退款原因
        :param refund_account: 可选，退款资金来源，仅针对老资金流商户使用，默认使用未结算资金退款
        :param notify_url: 可选，异步接收微信支付退款结果通知的回调地址
        :return: 返回的结果数据
        """
        data = {
            'appid': self.appid,
            'device_info': device_info,
            'transaction_id': transaction_id,
            'out_trade_no': out_trade_no,
            'out_refund_no': out_refund_no,
            'total_fee': total_fee,
            'refund_fee': refund_fee,
            'refund_fee_type': fee_type,
            'op_user_id': op_user_id if op_user_id else self.mch_id,
            'refund_account': refund_account,
            "refund_desc": refund_desc,
            'notify_url': notify_url,
        }
        return self._post('secapi/pay/refund', data=data)

    def query(self, refund_id=None, out_refund_no=None, transaction_id=None,
              out_trade_no=None, device_info=None):
        """
        查询退款

        :param refund_id: 微信退款单号
        :param out_refund_no: 商户退款单号
        :param transaction_id: 微信订单号
        :param out_trade_no: 商户系统内部的订单号
        :param device_info: 可选，终端设备号
        :return: 返回的结果数据
        """
        data = {
            'appid': self.appid,
            'device_info': device_info,
            'transaction_id': transaction_id,
            'out_trade_no': out_trade_no,
            'out_refund_no': out_refund_no,
            'refund_id': refund_id,
        }
        return self._post('pay/refundquery', data=data)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes


class BaseWeChatCipher(object):

    def __init__(self, cipher):
        self.cipher = cipher

    def encrypt(self, plaintext):
        encryptor = self.cipher.encryptor()
        return encryptor.update(plaintext) + encryptor.finalize()

    def decrypt(self, ciphertext):
        decryptor = self.cipher.decryptor()
        return decryptor.update(ciphertext) + decryptor.finalize()


class WeChatCipher(BaseWeChatCipher):

    def __init__(self, key, iv=None):
        iv = iv or key[:16]
        super(WeChatCipher, self).__init__(
            Cipher(
                algorithms.AES(key),
                modes.CBC(iv),
            )
        )


class AesEcbCipher(BaseWeChatCipher):

    def __init__(self, key):
        super(AesEcbCipher, self).__init__(
            Cipher(
                algorithms.AES(key),
                modes.ECB(),
            )
        )



# -*- coding: utf-8 -*-
"""
    wechatpyyyy.exceptions
    ~~~~~~~~~~~~~~~~~~~~

    Basic exceptions definition.

    :copyright: (c) 2014 by messense.
    :license: MIT, see LICENSE for more details.
"""
from __future__ import absolute_import, unicode_literals

import six

from wechatpyyyy.utils import to_binary, to_text


class WeChatException(Exception):
    """Base exception for wechatpy"""

    def __init__(self, errcode, errmsg):
        """
        :param errcode: Error code
        :param errmsg: Error message
        """
        self.errcode = errcode
        self.errmsg = errmsg

    def __str__(self):
        _repr = 'Error code: {code}, message: {msg}'.format(
            code=self.errcode,
            msg=self.errmsg
        )
        if six.PY2:
            return to_binary(_repr)
        else:
            return to_text(_repr)

    def __repr__(self):
        _repr = '{klass}({code}, {msg})'.format(
            klass=self.__class__.__name__,
            code=self.errcode,
            msg=self.errmsg
        )
        if six.PY2:
            return to_binary(_repr)
        else:
            return to_text(_repr)


class WeChatClientException(WeChatException):
    """WeChat API client exception class"""

    def __init__(self, errcode, errmsg, client=None,
                 request=None, response=None):
        super(WeChatClientException, self).__init__(errcode, errmsg)
        self.client = client
        self.request = request
        self.response = response


class InvalidSignatureException(WeChatException):
    """Invalid signature exception class"""

    def __init__(self, errcode=-40001, errmsg='Invalid signature'):
        super(InvalidSignatureException, self).__init__(errcode, errmsg)


class APILimitedException(WeChatClientException):
    """WeChat API call limited exception class"""
    pass


class InvalidAppIdException(WeChatException):
    """Invalid app_id exception class"""

    def __init__(self, errcode=-40005, errmsg='Invalid AppId'):
        super(InvalidAppIdException, self).__init__(errcode, errmsg)


class InvalidMchIdException(WeChatException):
    """Invalid mch_id exception class"""

    def __init__(self, errcode=-40006, errmsg='Invalid MchId'):
        super(InvalidMchIdException, self).__init__(errcode, errmsg)


class WeChatOAuthException(WeChatClientException):
    """WeChat OAuth API exception class"""
    pass


class WeChatComponentOAuthException(WeChatClientException):
    """WeChat Component OAuth API exception class"""
    pass


class WeChatPayException(WeChatClientException):
    """WeChat Pay API exception class"""

    def __init__(self, return_code, result_code=None, return_msg=None,
                 errcode=None, errmsg=None, client=None,
                 request=None, response=None):
        """
        :param return_code: 返回状态码
        :param result_code: 业务结果
        :param return_msg: 返回信息
        :param errcode: 错误代码
        :param errmsg: 错误代码描述
        """
        super(WeChatPayException, self).__init__(
            errcode,
            errmsg,
            client,
            request,
            response
        )
        self.return_code = return_code
        self.result_code = result_code
        self.return_msg = return_msg

    def __str__(self):
        _str = 'Error code: {code}, message: {msg}. Pay Error code: {pay_code}, message: {pay_msg}'.format(
                code=self.return_code,
                msg=self.return_msg,
                pay_code=self.errcode,
                pay_msg=self.errmsg
            )
        if six.PY2:
            return to_binary(_str)
        else:
            return to_text(_str)

    def __repr__(self):
        _repr = '{klass}({code}, {msg}). Pay({pay_code}, {pay_msg})'.format(
            klass=self.__class__.__name__,
            code=self.return_code,
            msg=self.return_msg,
            pay_code=self.errcode,
            pay_msg=self.errmsg
        )
        if six.PY2:
            return to_binary(_repr)
        else:
            return to_text(_repr)



# -*- coding: utf-8 -*-
"""
file: external_contact.py
author: Beck
created at: 2019-06-06
email: beck@ilife.co
"""
from __future__ import absolute_import, unicode_literals

from optionaldict import optionaldict

from wechatpyyyy.client.api.base import BaseWeChatAPI


class WeChatExternalContact(BaseWeChatAPI):
    """
    https://work.weixin.qq.com/api/doc#90000/90135/90221
    """

    def get_follow_user_list(self):
        """
        获取配置了客户联系功能的成员列表
        https://work.weixin.qq.com/api/doc#90000/90135/91554
        :return: 返回的 JSON 数据包
        """
        return self._get('externalcontact/get_follow_user_list')

    def list(self, userid):
        """
        获取外部联系人列表
        https://work.weixin.qq.com/api/doc#90000/90135/91555
        :param userid: 企业成员的userid
        :return: 返回的 JSON 数据包
        """
        return self._get('externalcontact/list', params={'userid': userid})

    def get(self, external_userid):
        """
        获取外部联系人详情
        https://work.weixin.qq.com/api/doc#90000/90135/91556
        :param external_userid: 外部联系人的userid，注意不是企业成员的帐号
        :return: 返回的 JSON 数据包
        """
        return self._get('externalcontact/get',
                         params={'external_userid': external_userid})

    def add_contact_way(self, type, scene, style=None,
                        remark=None, skip_verify=True, state=None,
                        user=None, party=None):
        """
        配置客户联系「联系我」方式
        https://work.weixin.qq.com/api/doc#90000/90135/91559

        :param type: 联系方式类型,1-单人, 2-多人
        :param scene: 场景，1-在小程序中联系，2-通过二维码联系
        :param style: 在小程序中联系时使用的控件样式，详见附表
        :param remark: 联系方式的备注信息，用于助记，不超过30个字符
        :param skip_verify: 外部客户添加时是否无需验证，默认为true
        :param state: 企业自定义的state参数，用于区分不同的添加渠道，在调用“获取外部联系人详情”时会返回该参数值
        :param user: 使用该联系方式的用户userID列表，在type为1时为必填，且只能有一个
        :param party: 使用该联系方式的部门id列表，只在type为2时有效
        :return: 返回的 JSON 数据包
        """
        data = optionaldict(
            type=type,
            scene=scene,
            style=style,
            remark=remark,
            skip_verify=skip_verify,
            state=state,
            user=user,
            party=party
        )
        return self._post('externalcontact/add_contact_way', data=data)

    def get_contact_way(self, config_id):
        """
        获取企业已配置的「联系我」方式
        https://work.weixin.qq.com/api/doc#90000/90135/91559
        :param config_id: 联系方式的配置id, e.g.42b34949e138eb6e027c123cba77fad7
        :return: 返回的 JSON 数据包
        """
        data = optionaldict(
            config_id=config_id
        )
        return self._post('externalcontact/get_contact_way', data=data)

    def update_contact_way(self, config_id, remark, skip_verify=True,
                           style=None, state=None, user=None,
                           party=None):
        """
        更新企业已配置的「联系我」方式
        https://work.weixin.qq.com/api/doc#90000/90135/91559

        :param config_id: 企业联系方式的配置id
        :param remark: 联系方式的备注信息，不超过30个字符，将覆盖之前的备注
        :param skip_verify: 外部客户添加时是否无需验证
        :param style: 样式，只针对“在小程序中联系”的配置生效
        :param state: 企业自定义的state参数，用于区分不同的添加渠道，在调用“获取外部联系人详情”时会返回该参数值
        :param user: 使用该联系方式的用户列表，将覆盖原有用户列表
        :param party: 使用该联系方式的部门列表，将覆盖原有部门列表，只在配置的type为2时有效
        :return: 返回的 JSON 数据包
        """
        data = optionaldict(
            config_id=config_id,
            remark=remark,
            skip_verify=skip_verify,
            style=style,
            state=state,
            user=user,
            party=party
        )
        return self._post('externalcontact/update_contact_way', data=data)

    def del_contact_way(self, config_id):
        """
        删除企业已配置的「联系我」方式
        :param config_id: 企业联系方式的配置id
        :return: 返回的 JSON 数据包
        """
        data = optionaldict(
            config_id=config_id
        )
        return self._post('externalcontact/del_contact_way', data=data)

    def add_msg_template(self, template):
        """
        添加企业群发消息模板
        https://work.weixin.qq.com/api/doc#90000/90135/91560

        {
          "external_userid":[
            "woAJ2GCAAAXtWyujaWJHDDGi0mACas1w",
            "wmqfasd1e1927831291723123109r712"
          ],
          "sender":"zhangsan",
          "text":{
            "content":"文本消息内容"
          },
          "image":{
            "media_id":"MEDIA_ID"
          },
          "link":{
            "title":"消息标题",
            "picurl":"https://example.pic.com/path",
            "desc":"消息描述",
            "url":"https://example.link.com/path"
          },
          "miniprogram":{
            "title":"消息标题",
            "pic_media_id":"MEDIA_ID",
            "appid":"wx8bd80126147df384",
            "page":"/path/index"
          }
        }
        external_userid	否	客户的外部联系人id列表，不可与sender同时为空，最多可传入1万个客户
        sender	否	发送企业群发消息的成员userid，不可与external_userid同时为空
        text.content	否	消息文本内容
        image.media_id	是	图片的media_id
        link.title	是	图文消息标题
        link.picurl	否	图文消息封面的url
        link.desc	否	图文消息的描述
        link.url	是	图文消息的链接
        miniprogram.title	是	小程序消息标题
        miniprogram.pic_media_id	是	小程序消息封面的mediaid，封面图建议尺寸为520*416
        miniprogram.appid	是	小程序appid，必须是关联到企业的小程序应用
        miniprogram.page	是	小程序page路径

        text、image、link和miniprogram四者不能同时为空；
        text与另外三者可以同时发送，此时将会以两条消息的形式触达客户
        image、link和miniprogram只能有一个，如果三者同时填，则按image、link、miniprogram的优先顺序取参，也就是说，如果image与link同时传值，则只有image生效。
        media_id可以通过素材管理接口获得。

        :param template: 见上方说明.
        :return: 返回的 JSON 数据包
        """
        return self._post('externalcontact/add_msg_template', data=template)

    def get_group_msg_result(self, msgid):
        """
        获取企业群发消息发送结果
        企业和第三方可通过该接口获取到添加企业群发消息模板生成消息的群发发送结果。
        https://work.weixin.qq.com/api/doc#90000/90135/91561

        :param msgid: 群发消息的id，通过添加企业群发消息模板接口返回
        :return: 返回的 JSON 数据包
        """
        data = optionaldict(
            msgid=msgid
        )
        return self._post('externalcontact/get_group_msg_result', data=data)

    def get_user_behavior_data(self, userid,
                               start_time, end_time):
        """
        获取员工行为数据

        企业可通过此接口获取员工联系客户的行为数据，包括聊天数，发送消息数，消息回复比例和平均首次回复时长等维度。
        https://work.weixin.qq.com/api/doc#90000/90135/91580

        :param userid: 	userid列表
        :param start_time: 数据起始时间
        :param end_time: 数据结束时间
        :return: 返回的 JSON 数据包
        """
        data = optionaldict(
            userid=userid,
            start_time=start_time,
            end_time=end_time
        )
        return self._post('externalcontact/get_user_behavior_data', data=data)

    def send_welcome_msg(self, template):
        """
        发送新客户欢迎语
        企业微信在向企业推送添加外部联系人事件时，会额外返回一个welcome_code，企业以此为凭据调用接口，即可通过成员向新添加的客户发送个性化的欢迎语。
        为了保证用户体验以及避免滥用，企业仅可在收到相关事件后20秒内调用，且只可调用一次。
        如果企业已经在管理端为相关成员配置了可用的欢迎语，则推送添加外部联系人事件时不会返回welcome_code。
        https://work.weixin.qq.com/api/doc#90000/90135/91688

        {
          "welcome_code":"CALLBACK_CODE",
          "text":{
            "content":"文本消息内容"
          },
          "image":{
            "media_id":"MEDIA_ID"
          },
          "link":{
            "title":"消息标题",
            "picurl":"https://example.pic.com/path",
            "desc":"消息描述",
            "url":"https://example.link.com/path"
          },
          "miniprogram":{
            "title":"消息标题",
            "pic_media_id":"MEDIA_ID",
            "appid":"wx8bd80126147df384",
            "page":"/path/index"
          }
        }
        welcome_code	是	通过添加外部联系人事件推送给企业的发送欢迎语的凭证，有效期为20秒
        text.content	否	消息文本内容
        image.media_id	是	图片的media_id
        link.title	是	图文消息标题
        link.picurl	否	图文消息封面的url
        link.desc	否	图文消息的描述
        link.url	是	图文消息的链接
        miniprogram.title	是	小程序消息标题
        miniprogram.pic_media_id	是	小程序消息封面的mediaid，封面图建议尺寸为520*416
        miniprogram.appid	是	小程序appid，必须是关联到企业的小程序应用
        miniprogram.page	是	小程序page路径

        text、image、link和miniprogram四者不能同时为空；
        text与另外三者可以同时发送，此时将会以两条消息的形式触达客户
        image、link和miniprogram只能有一个，如果三者同时填，则按image、link、miniprogram的优先顺序取参，也就是说，如果image与link同时传值，则只有image生效。
        media_id可以通过素材管理接口获得。
        :param template: 见上方说明.
        :return: 返回的 JSON 数据包
        """
        return self._post('externalcontact/send_welcome_msg', data=template)

    def get_unassigned_list(self, page_id, page_size):
        """
        获取离职成员的客户列表
        企业和第三方可通过此接口，获取所有离职成员的客户列表，并可进一步调用离职成员的外部联系人再分配接口将这些客户重新分配给其他企业成员。
        https://work.weixin.qq.com/api/doc#90000/90135/91563

        :param page_id: 分页查询，要查询页号，从0开始
        :param page_size: 每次返回的最大记录数，默认为1000，最大值为1000
        :return:
        """
        data = optionaldict(
            page_id=page_id,
            page_size=page_size
        )
        return self._post('externalcontact/get_unassigned_list', data=data)

    def transfer(self, external_userid, handover_userid,
                 takeover_userid):
        """
        离职成员的外部联系人再分配
        企业可通过此接口，将已离职成员的外部联系人分配给另一个成员接替联系。
        https://work.weixin.qq.com/api/doc#90000/90135/91564

        :param external_userid: 外部联系人的userid，注意不是企业成员的帐号
        :param handover_userid: 离职成员的userid
        :param takeover_userid: 接替成员的userid
        :return: 返回的 JSON 数据包
        """
        data = optionaldict(
            external_userid=external_userid,
            handover_userid=handover_userid,
            takeover_userid=takeover_userid
        )
        return self._post('externalcontact/transfer', data=data)

    def get_corp_tag_list(self, tag_ids=None):
        """
        企业可通过此接口获取企业客户标签详情
        https://work.weixin.qq.com/api/doc/90000/90135/92116

        :param tag_ids:要查询的标签id，如果不填则获取该企业的所有客户标签，目前暂不支持标签组id ,示例：["etXXXXXXXXXX","etYYYYYYYYYY"]
        :return: 返回的 JSON 数据包
        """
        data = optionaldict(
            tag_id=tag_ids
        )
        return self._post('externalcontact/get_corp_tag_list', data=data)

    def add_corp_tag(self, group_id, group_name, order, tags):
        """

        企业可通过此接口向客户标签库中添加新的标签组和标签。
        https://work.weixin.qq.com/api/doc/90000/90135/92117

        :param group_id: 标签组id
        :param group_name: 标签组名称，最长为30个字符
        :param order: 否	标签组次序值。order值大的排序靠前。有效的值范围是[0, 2^32)
        :param tags:[
                    tag.name	是	添加的标签名称，最长为30个字符
                    tag.order	否	标签次序值。order值大的排序靠前。有效的值范围是[0, 2^32)
                    ]
        :return:返回的 JSON 数据包
        """
        data = optionaldict(
            group_id=group_id,
            group_name=group_name,
            order=order,
            tag=tags
        )

        return self._post('externalcontact/add_corp_tag', data=data)

    def edit_corp_tag(self, id, name, order):
        """
        企业可通过此接口编辑客户标签/标签组的名称或次序值。
        https://work.weixin.qq.com/api/doc/90000/90135/92117

        :param id:标签或标签组的id列表
        :param name:新的标签或标签组名称，最长为30个字符
        :param order:标签/标签组的次序值。order值大的排序靠前。有效的值范围是[0, 2^32)
        :return:返回的 JSON 数据包
        """
        data = optionaldict(
            id=id,
            name=name,
            order=order
        )
        return self._post('externalcontact/edit_corp_tag', data=data)

    def del_corp_tag(self, tag_id=None, group_id=None):
        """
        企业可通过此接口删除客户标签库中的标签，或删除整个标签组。
        如果一个标签组下所有的标签均被删除，则标签组会被自动删除。
        https://work.weixin.qq.com/api/doc/90000/90135/92117

        :param tag_id:标签的id列表  tag_id和group_id不可同时为空。
        :param group_id:标签组的id列表  tag_id和group_id不可同时为空。
        :return:返回的 JSON 数据包
        """
        data = optionaldict(
            tag_id=tag_id,
            group_id=group_id
        )
        return self._post('externalcontact/del_corp_tag', data=data)

    def mark_tag(self, userid, external_userid, add_tag=None, remove_tag=None):
        """
        企业可通过此接口为指定成员的客户添加上由企业统一配置的标签。
        https://work.weixin.qq.com/api/doc/90000/90135/92118

        :param userid:添加外部联系人的userid
        :param external_userid:外部联系人userid
        :param add_tag:要标记的标签列表
        :param remove_tag:要移除的标签列表
        :return:返回的 JSON 数据包
        """
        add_tag = add_tag or []
        remove_tag = remove_tag or []
        data = optionaldict(
            userid=userid,
            external_userid=external_userid,
            add_tag=add_tag,
            remove_tag=remove_tag
        )
        return self._post('externalcontact/mark_tag', data=data)



from __future__ import absolute_import, unicode_literals

import logging

from wechatpyyyy.client import WeChatClient  # NOQA
from wechatpyyyy.component import ComponentOAuth, WeChatComponent  # NOQA
from wechatpyyyy.exceptions import WeChatClientException, WeChatException, WeChatOAuthException, WeChatPayException  # NOQA
from wechatpyyyy.oauth import WeChatOAuth  # NOQA
from wechatpyyyy.parser import parse_message  # NOQA
from wechatpyyyy.pay import WeChatPay  # NOQA
from wechatpyyyy.replies import create_reply  # NOQA

__version__ = '1.8.18'
__author__ = 'messense'

# Set default logging handler to avoid "No handler found" warnings.
try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

logging.getLogger(__name__).addHandler(NullHandler())



# -*- coding: utf-8 -*-
"""
    wechatpyyyy.fields
    ~~~~~~~~~~~~~~~~

    This module defines some useful field types for parse WeChat messages

    :copyright: (c) 2014 by messense.
    :license: MIT, see LICENSE for more details.
"""
from __future__ import absolute_import, unicode_literals
import time
from datetime import datetime
import base64
import copy

import six

from wechatpyyyy.utils import to_text, to_binary, ObjectDict, timezone


default_timezone = timezone('Asia/Shanghai')


class FieldDescriptor(object):

    def __init__(self, field):
        self.field = field
        self.attr_name = field.name

    def __get__(self, instance, instance_type=None):
        if instance is not None:
            value = instance._data.get(self.attr_name)
            if value is None:
                value = copy.deepcopy(self.field.default)
                instance._data[self.attr_name] = value
            if isinstance(value, dict):
                value = ObjectDict(value)
            if value and not isinstance(value, (dict, list, tuple)) and \
                    six.callable(self.field.converter):
                value = self.field.converter(value)
            return value
        return self.field

    def __set__(self, instance, value):
        instance._data[self.attr_name] = value


class BaseField(object):
    converter = None

    def __init__(self, name, default=None):
        self.name = name
        self.default = default

    def to_xml(self, value):
        raise NotImplementedError()

    @classmethod
    def from_xml(cls, value):
        raise NotImplementedError()

    def __repr__(self):
        _repr = '{klass}({name})'.format(
            klass=self.__class__.__name__,
            name=repr(self.name)
        )
        if six.PY2:
            return to_binary(_repr)
        else:
            return to_text(_repr)

    def add_to_class(self, klass, name):
        self.klass = klass
        klass._fields[name] = self
        setattr(klass, name, FieldDescriptor(self))


class StringField(BaseField):

    def __to_text(self, value):
        return to_text(value)

    converter = __to_text

    def to_xml(self, value):
        value = self.converter(value)
        tpl = '<{name}><![CDATA[{value}]]></{name}>'
        return tpl.format(name=self.name, value=value)

    @classmethod
    def from_xml(cls, value):
        return value


class IntegerField(BaseField):
    converter = int

    def to_xml(self, value):
        value = self.converter(value) if value is not None else self.default
        tpl = '<{name}>{value}</{name}>'
        return tpl.format(name=self.name, value=value)

    @classmethod
    def from_xml(cls, value):
        return cls.converter(value)


class DateTimeField(BaseField):
    def __converter(self, value):
        v = int(value)
        return datetime.fromtimestamp(v, tz=default_timezone)
    converter = __converter

    def to_xml(self, value):
        value = time.mktime(datetime.timetuple(value))
        value = int(value)
        tpl = '<{name}>{value}</{name}>'
        return tpl.format(name=self.name, value=value)

    @classmethod
    def from_xml(cls, value):
        return cls.converter(None, value)


class FloatField(BaseField):
    converter = float

    def to_xml(self, value):
        value = self.converter(value) if value is not None else self.default
        tpl = '<{name}>{value}</{name}>'
        return tpl.format(name=self.name, value=value)

    @classmethod
    def from_xml(cls, value):
        return cls.converter(value)


class ImageField(StringField):

    def to_xml(self, value):
        value = self.converter(value)
        tpl = """<Image>
        <MediaId><![CDATA[{value}]]></MediaId>
        </Image>"""
        return tpl.format(value=value)

    @classmethod
    def from_xml(cls, value):
        return value["MediaId"]


class VoiceField(StringField):

    def to_xml(self, value):
        value = self.converter(value)
        tpl = """<Voice>
        <MediaId><![CDATA[{value}]]></MediaId>
        </Voice>"""
        return tpl.format(value=value)

    @classmethod
    def from_xml(cls, value):
        return value["MediaId"]


class VideoField(StringField):

    def to_xml(self, value):
        kwargs = dict(media_id=self.converter(value['media_id']))
        content = '<MediaId><![CDATA[{media_id}]]></MediaId>'
        if 'title' in value:
            kwargs['title'] = self.converter(value['title'])
            content += '<Title><![CDATA[{title}]]></Title>'
        if 'description' in value:
            kwargs['description'] = self.converter(value['description'])
            content += '<Description><![CDATA[{description}]]></Description>'
        tpl = """<Video>
        {content}
        </Video>""".format(content=content)
        return tpl.format(**kwargs)

    @classmethod
    def from_xml(cls, value):
        rv = dict(media_id=value['MediaId'])
        if 'Title' in value:
            rv["title"] = value['Title']
        if 'Description' in value:
            rv['description'] = value['Description']
        return rv


class MusicField(StringField):

    def to_xml(self, value):
        kwargs = dict(thumb_media_id=self.converter(value['thumb_media_id']))
        content = '<ThumbMediaId><![CDATA[{thumb_media_id}]]></ThumbMediaId>'
        if 'title' in value:
            kwargs['title'] = self.converter(value['title'])
            content += '<Title><![CDATA[{title}]]></Title>'
        if 'description' in value:
            kwargs['description'] = self.converter(value['description'])
            content += '<Description><![CDATA[{description}]]></Description>'
        if 'music_url' in value:
            kwargs['music_url'] = self.converter(value['music_url'])
            content += '<MusicUrl><![CDATA[{music_url}]]></MusicUrl>'
        if 'hq_music_url' in value:
            kwargs['hq_music_url'] = self.converter(value['hq_music_url'])
            content += '<HQMusicUrl><![CDATA[{hq_music_url}]]></HQMusicUrl>'
        tpl = """<Music>
        {content}
        </Music>""".format(content=content)
        return tpl.format(**kwargs)

    @classmethod
    def from_xml(cls, value):
        rv = dict(thumb_media_id=value['ThumbMediaId'])
        if 'Title' in value:
            rv['title'] = value['Title']
        if 'Description' in value:
            rv['description'] = value['Description']
        if 'MusicUrl' in value:
            rv['music_url'] = value['MusicUrl']
        if 'HQMusicUrl' in value:
            rv['hq_music_url'] = value['HQMusicUrl']
        return rv


class ArticlesField(StringField):

    def to_xml(self, articles):
        article_count = len(articles)
        items = []
        for article in articles:
            title = self.converter(article.get('title', ''))
            description = self.converter(article.get('description', ''))
            image = self.converter(article.get('image', ''))
            url = self.converter(article.get('url', ''))
            item_tpl = """<item>
            <Title><![CDATA[{title}]]></Title>
            <Description><![CDATA[{description}]]></Description>
            <PicUrl><![CDATA[{image}]]></PicUrl>
            <Url><![CDATA[{url}]]></Url>
            </item>"""
            item = item_tpl.format(
                title=title,
                description=description,
                image=image,
                url=url
            )
            items.append(item)
        items_str = '\n'.join(items)
        tpl = """<ArticleCount>{article_count}</ArticleCount>
        <Articles>{items}</Articles>"""
        return tpl.format(
            article_count=article_count,
            items=items_str
        )

    @classmethod
    def from_xml(cls, value):
        return [dict(
            title=item["Title"],
            description=item["Description"],
            image=item["PicUrl"],
            url=item["Url"]
        ) for item in value["item"]]


class Base64EncodeField(StringField):

    def __base64_encode(self, text):
        return to_text(base64.b64encode(to_binary(text)))

    converter = __base64_encode


class Base64DecodeField(StringField):

    def __base64_decode(self, text):
        return to_text(base64.b64decode(to_binary(text)))

    converter = __base64_decode


class HardwareField(StringField):

    def to_xml(self, value=None):
        value = value or {'view': 'myrank', 'action': 'ranklist'}
        tpl = """<{name}>
        <MessageView><![CDATA[{view}]]></MessageView>
        <MessageAction><![CDATA[{action}]]></MessageAction>
        </{name}>"""
        return tpl.format(
            name=self.name,
            view=value.get('view'),
            action=value.get('action')
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from wechatpyyyy.client.api.base import BaseWeChatAPI


class WeChatMenu(BaseWeChatAPI):
    """
    自定义菜单

    https://work.weixin.qq.com/api/doc#90000/90135/90230
    """

    def create(self, agent_id, menu_data):
        """
        创建菜单

        https://work.weixin.qq.com/api/doc#90000/90135/90231

        :param agent_id: 应用id
        """
        return self._post(
            'menu/create',
            params={
                'agentid': agent_id
            },
            data=menu_data
        )

    def get(self, agent_id):
        """
        获取菜单

        https://work.weixin.qq.com/api/doc#90000/90135/90232

        :param agent_id: 应用id
        """
        return self._get(
            'menu/get',
            params={
                'agentid': agent_id
            }
        )

    def delete(self, agent_id):
        """
        删除菜单

        https://work.weixin.qq.com/api/doc#90000/90135/90233

        :param agent_id: 应用id
        """
        return self._get(
            'menu/delete',
            params={
                'agentid': agent_id
            }
        )

    def update(self, agent_id, menu_data):
        self.delete(agent_id)
        return self.create(agent_id, menu_data)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from optionaldict import optionaldict
from wechatpyyyy.client.api.base import BaseWeChatAPI


class WeChatOA(BaseWeChatAPI):
    """
    OA管理

    https://work.weixin.qq.com/api/doc/90000/90135/90264
    """

    def get_template_detail(self, template_id):
        """
        查询审批模板的详情
        https://work.weixin.qq.com/api/doc/90000/90135/91982

        :param template_id: 模板Id
        :return:
        """
        data = {
            'template_id': template_id
        }
        return self._post(
            'oa/gettemplatedetail',
            data=data
        )

    def get_approval_info(self, start_time, end_time, cursor, size=100, filters=None):
        """
        批量获取审批单号
        https://work.weixin.qq.com/api/doc/90000/90135/91816

        :param start_time: 开始时间戳
        :param end_time: 结束时间戳，请求的参数endtime需要大于startime， 起始时间跨度不能超过31天；
        :param cursor: 分页查询游标，默认为0，后续使用返回的next_cursor进行分页拉取
        :param size: 一次请求拉取审批单数量，默认值为100，上限值为100
        :param filters: 请自行查看文档
        :return:
        """
        data = optionaldict({
            'starttime': str(start_time),
            'endtime': str(end_time),
            'cursor': cursor,
            'size': size,
            'filter': filters
        })

        return self._post(
            'oa/getapprovalinfo',
            data=data
        )

    def get_approval_detail(self, sp_no):
        """
        获取审批申请详情
        https://work.weixin.qq.com/api/doc/90000/90135/91983

        :param sp_no: 审批单编号
        :return:
        """
        data = {
            'sp_no': sp_no
        }
        return self._post(
            'oa/getapprovaldetail',
            data=data
        )

    def apply_event(self, creator_userid, template_id, use_template_approver, approver, apply_data, summary_list,
                    notifyer=None, notify_type=None):
        """
        提交审批申请，这个函数的参数比较复杂，具体请查看官方文档
        https://work.weixin.qq.com/api/doc/90000/90135/91853

        :param creator_userid: 申请人userid，此审批申请将以此员工身份提交，申请人需在应用可见范围内
        :param template_id: 模板id。可在“获取审批申请详情”、“审批状态变化回调通知”中获得，也可在审批模板的模板编辑页面链接中获得。暂不支持通过接口提交[打卡补卡][调班]模板审批单。
        :param use_template_approver: 审批人模式：0-通过接口指定审批人、抄送人（此时approver、notifyer等参数可用）; 1-使用此模板在管理后台设置的审批流程，支持条件审批。
        :param approver: 具体参数查看官方文档，审批流程信息，用于指定审批申请的审批流程，支持单人审批、多人会签、多人或签，可能有多个审批节点，仅use_template_approver为0时生效。
        :param apply_data: 具体参数查看官方文档，审批申请数据，可定义审批申请中各个控件的值，其中必填项必须有值，选填项可为空，数据结构同“获取审批申请详情”接口返回值中同名参数“apply_data”
        :param summary_list: 具体参数查看官方文档，摘要信息，用于显示在审批通知卡片、审批列表的摘要信息，最多3行
        :param notifyer: 抄送人节点userid列表，仅use_template_approver为0时生效。
        :param notify_type: 抄送方式：1-提单时抄送（默认值）； 2-单据通过后抄送；3-提单和单据通过后抄送。仅use_template_approver为0时生效。
        :return:
        """
        data = optionaldict({
            'creator_userid': creator_userid,
            'template_id': template_id,
            'use_template_approver': use_template_approver,
            'approver': approver,
            'notifyer': notifyer,
            'notify_type': notify_type,
            'apply_data': apply_data,
            'summary_list': summary_list
        })
        return self._post(
            'oa/applyevent',
            data=data
        )


# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import time
import random
from datetime import datetime, timedelta

from wechatpyyyy.utils import timezone
from wechatpyyyy.pay.utils import get_external_ip
from wechatpyyyy.pay.base import BaseWeChatPayAPI
from wechatpyyyy.utils import random_string, to_text, json
from wechatpyyyy.pay.utils import calculate_signature


class WeChatOrder(BaseWeChatPayAPI):

    def create(self, trade_type, body, total_fee, notify_url, client_ip=None,
               user_id=None, out_trade_no=None, detail=None, attach=None,
               fee_type='CNY', time_start=None, time_expire=None,
               goods_tag=None, product_id=None, device_info=None,
               limit_pay=None, scene_info=None, sub_user_id=None, **kwargs):
        """
        统一下单接口

        :param trade_type: 交易类型，取值如下：JSAPI，NATIVE，APP，WAP, MWEB
        :param body: 商品描述
        :param total_fee: 总金额，单位分
        :param notify_url: 接收微信支付异步通知回调地址
        :param client_ip: 可选，APP和网页支付提交用户端ip，Native支付填调用微信支付API的机器IP
        :param user_id: 可选，用户在商户appid下的唯一标识。trade_type=JSAPI和appid已设定，此参数必传
        :param sub_user_id: 可选，小程序appid下的唯一标识。trade_type=JSAPI和sub_appid已设定，此参数必传
        :param out_trade_no: 可选，商户订单号，默认自动生成
        :param detail: 可选，商品详情
        :param attach: 可选，附加数据，在查询API和支付通知中原样返回，该字段主要用于商户携带订单的自定义数据
        :param fee_type: 可选，符合ISO 4217标准的三位字母代码，默认人民币：CNY
        :param time_start: 可选，订单生成时间，默认为当前时间
        :param time_expire: 可选，订单失效时间，默认为订单生成时间后两小时
        :param goods_tag: 可选，商品标记，代金券或立减优惠功能的参数
        :param product_id: 可选，trade_type=NATIVE，此参数必传。此id为二维码中包含的商品ID，商户自行定义
        :param device_info: 可选，终端设备号(门店号或收银设备ID)，注意：PC网页或公众号内支付请传"WEB"
        :param limit_pay: 可选，指定支付方式，no_credit--指定不能使用信用卡支付
        :param scene_info: 可选，上报支付的场景信息
        :param kwargs: 其他未列举在上述参数中的统一下单接口调用参数,例如电子发票入口开放标识receipt
        :type scene_info: dict
        :return: 返回的结果数据
        """
        now = datetime.fromtimestamp(time.time(), tz=timezone('Asia/Shanghai'))
        hours_later = now + timedelta(hours=2)
        if time_start is None:
            time_start = now
        if time_expire is None:
            time_expire = hours_later
        if not out_trade_no:
            out_trade_no = '{0}{1}{2}'.format(
                self.mch_id,
                now.strftime('%Y%m%d%H%M%S'),
                random.randint(1000, 10000)
            )
        if scene_info is not None:
            scene_info = json.dumps(scene_info, ensure_ascii=False)
        data = {
            'appid': self.appid,
            'sub_appid': self.sub_appid,
            'device_info': device_info,
            'body': body,
            'detail': detail,
            'attach': attach,
            'out_trade_no': out_trade_no,
            'fee_type': fee_type,
            'total_fee': total_fee,
            'spbill_create_ip': client_ip or get_external_ip(),
            'time_start': time_start.strftime('%Y%m%d%H%M%S'),
            'time_expire': time_expire.strftime('%Y%m%d%H%M%S'),
            'goods_tag': goods_tag,
            'notify_url': notify_url,
            'trade_type': trade_type,
            'limit_pay': limit_pay,
            'product_id': product_id,
            'openid': user_id,
            'sub_openid': sub_user_id,
            'scene_info': scene_info,
        }
        data.update(kwargs)
        return self._post('pay/unifiedorder', data=data)

    def query(self, transaction_id=None, out_trade_no=None):
        """
        查询订单

        :param transaction_id: 微信的订单号，优先使用
        :param out_trade_no: 商户系统内部的订单号，当没提供transaction_id时需要传这个。
        :return: 返回的结果数据
        """
        data = {
            'appid': self.appid,
            'transaction_id': transaction_id,
            'out_trade_no': out_trade_no,
        }
        return self._post('pay/orderquery', data=data)

    def close(self, out_trade_no):
        """
        关闭订单

        :param out_trade_no: 商户系统内部的订单号
        :return: 返回的结果数据
        """
        data = {
            'appid': self.appid,
            'out_trade_no': out_trade_no,
        }
        return self._post('pay/closeorder', data=data)

    def get_appapi_params(self, prepay_id, timestamp=None, nonce_str=None):
        """
        获取 APP 支付参数

        :param prepay_id: 统一下单接口返回的 prepay_id 参数值
        :param timestamp: 可选，时间戳，默认为当前时间戳
        :param nonce_str: 可选，随机字符串，默认自动生成
        :return: 签名
        """
        data = {
            'appid': self.appid,
            'partnerid': self.mch_id,
            'prepayid': prepay_id,
            'package': 'Sign=WXPay',
            'timestamp': timestamp or to_text(int(time.time())),
            'noncestr': nonce_str or random_string(32)
        }
        sign = calculate_signature(data, self._client.api_key)
        data['sign'] = sign
        return data

    def reverse(self, transaction_id=None, out_trade_no=None):
        """
        撤销订单

        :param transaction_id: 可选，微信的订单号，优先使用
        :param out_trade_no: 可选，商户系统内部的订单号,
                            transaction_id、out_trade_no二选一，
                            如果同时存在优先级：transaction_id> out_trade_no
        :return: 返回的结果数据
        """
        data = {
            'appid': self.appid,
            'transaction_id': transaction_id,
            'out_trade_no': out_trade_no,
        }
        return self._post('secapi/pay/reverse', data=data)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import base64
import copy
import hashlib
import hmac
import socket
import logging
import six

from wechatpyyyy.utils import to_binary, to_text

logger = logging.getLogger(__name__)


def format_url(params, api_key=None):
    data = [to_binary('{0}={1}'.format(k, params[k])) for k in sorted(params) if params[k]]
    if api_key:
        data.append(to_binary('key={0}'.format(api_key)))
    return b"&".join(data)


def calculate_signature(params, api_key):
    url = format_url(params, api_key)
    logger.debug("Calculate Signature URL: %s", url)
    return to_text(hashlib.md5(url).hexdigest().upper())


def calculate_signature_hmac(params, api_key):
    url = format_url(params, api_key)
    sign = to_text(hmac.new(api_key.encode(), msg=url,
                            digestmod=hashlib.sha256).hexdigest().upper())
    return sign


def _check_signature(params, api_key):
    _params = copy.deepcopy(params)
    sign = _params.pop('sign', '')
    return sign == calculate_signature(_params, api_key)


def dict_to_xml(d, sign=None):
    xml = ['<xml>\n']
    for k in sorted(d):
        # use sorted to avoid test error on Py3k
        v = d[k]
        if isinstance(v, six.integer_types) or (isinstance(v, six.string_types) and v.isdigit()):
            xml.append('<{0}>{1}</{0}>\n'.format(to_text(k), to_text(v)))
        else:
            xml.append(
                '<{0}><![CDATA[{1}]]></{0}>\n'.format(to_text(k), to_text(v))
            )
    if sign:
        xml.append('<sign><![CDATA[{0}]]></sign>\n</xml>'.format(to_text(sign)))
    else:
        xml.append('</xml>')
    return ''.join(xml)


def get_external_ip():
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        wechat_ip = socket.gethostbyname('api.mch.weixin.qq.com')
        sock.connect((wechat_ip, 80))
        addr, port = sock.getsockname()
        sock.close()
        return addr
    except socket.error:
        return '127.0.0.1'


def rsa_encrypt(data, pem, b64_encode=True):
    """
    rsa 加密
    :param data: 待加密字符串/binary
    :param pem: RSA public key 内容/binary
    :param b64_encode: 是否对输出进行 base64 encode
    :return: 如果 b64_encode=True 的话，返回加密并 base64 处理后的 string；否则返回加密后的 binary
    """
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.primitives import hashes
    from cryptography.hazmat.primitives.asymmetric import padding
    encoded_data = to_binary(data)
    pem = to_binary(pem)
    public_key = serialization.load_pem_public_key(pem)
    encrypted_data = public_key.encrypt(
        encoded_data,
        padding=padding.OAEP(
            mgf=padding.MGF1(hashes.SHA1()),
            algorithm=hashes.SHA1(),
            label=None,
        )
    )
    if b64_encode:
        encrypted_data = base64.b64encode(encrypted_data).decode('utf-8')
    return encrypted_data


def rsa_decrypt(encrypted_data, pem, password=None):
    """
    rsa 解密
    :param encrypted_data: 待解密 bytes
    :param pem: RSA private key 内容/binary
    :param password: RSA private key pass phrase
    :return: 解密后的 binary
    """
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.primitives import hashes
    from cryptography.hazmat.primitives.asymmetric import padding

    encrypted_data = to_binary(encrypted_data)
    pem = to_binary(pem)
    private_key = serialization.load_pem_private_key(pem, password)
    data = private_key.decrypt(
        encrypted_data,
        padding=padding.OAEP(
            mgf=padding.MGF1(hashes.SHA1()),
            algorithm=hashes.SHA1(),
            label=None,
        )
    )
    return data



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
from wechatpyyyy.client.api.base import BaseWeChatAPI


class MerchantShelf(BaseWeChatAPI):

    API_BASE_URL = 'https://api.weixin.qq.com/'

    def add(self, name, banner, shelf_data):
        return self._post(
            'merchant/shelf/add',
            data={
                'shelf_name': name,
                'shelf_banner': banner,
                'shelf_data': shelf_data
            }
        )

    def delete(self, shelf_id):
        return self._post(
            'merchant/shelf/del',
            data={
                'shelf_id': shelf_id
            }
        )

    def update(self, shelf_id, name, banner, shelf_data):
        return self._post(
            'merchant/shelf/add',
            data={
                'shelf_id': shelf_id,
                'shelf_name': name,
                'shelf_banner': banner,
                'shelf_data': shelf_data
            }
        )

    def get_all(self):
        res = self._get(
            'merchant/shelf/getall',
            result_processor=lambda x: x['shelves']
        )
        return res

    def get(self, shelf_id):
        return self._post(
            'merchant/shelf/getbyid',
            data={
                'shelf_id': shelf_id
            }
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import random
from datetime import datetime

from wechatpyyyy.pay.utils import get_external_ip, rsa_encrypt
from wechatpyyyy.pay.base import BaseWeChatPayAPI


class WeChatTransfer(BaseWeChatPayAPI):

    def transfer(self, user_id, amount, desc, client_ip=None,
                 check_name='OPTION_CHECK', real_name=None,
                 out_trade_no=None, device_info=None):
        """
        企业付款接口

        :param user_id: 接受收红包的用户在公众号下的 openid
        :param amount: 付款金额，单位分
        :param desc: 付款说明
        :param client_ip: 可选，调用接口机器的 IP 地址
        :param check_name: 可选，校验用户姓名选项，
                           NO_CHECK：不校验真实姓名,
                           FORCE_CHECK：强校验真实姓名（未实名认证的用户会校验失败，无法转账）,
                           OPTION_CHECK：针对已实名认证的用户才校验真实姓名（未实名认证用户不校验，可以转账成功）,
                           默认为 OPTION_CHECK
        :param real_name: 可选，收款用户真实姓名，
                          如果check_name设置为FORCE_CHECK或OPTION_CHECK，则必填用户真实姓名
        :param out_trade_no: 可选，商户订单号，需保持唯一性，默认自动生成
        :param device_info: 可选，微信支付分配的终端设备号
        :return: 返回的结果信息
        """
        if not out_trade_no:
            now = datetime.now()
            out_trade_no = '{0}{1}{2}'.format(
                self.mch_id,
                now.strftime('%Y%m%d%H%M%S'),
                random.randint(1000, 10000)
            )
        data = {
            'mch_appid': self.appid,
            'mchid': self.mch_id,
            'device_info': device_info,
            'partner_trade_no': out_trade_no,
            'openid': user_id,
            'check_name': check_name,
            're_user_name': real_name,
            'amount': amount,
            'desc': desc,
            'spbill_create_ip': client_ip or get_external_ip(),
        }
        return self._post('mmpaymkttransfers/promotion/transfers', data=data)

    def query(self, out_trade_no):
        """
        企业付款查询接口

        :param out_trade_no: 商户调用企业付款API时使用的商户订单号
        :return: 返回的结果数据
        """
        data = {
            'appid': self.appid,
            'partner_trade_no': out_trade_no,
        }
        return self._post('mmpaymkttransfers/gettransferinfo', data=data)

    def transfer_bankcard(self, true_name, bank_card_no, bank_code, amount, desc=None, out_trade_no=None):
        """
        企业付款到银行卡接口

        :param true_name: 开户人名称
        :param bank_card_no: 银行卡号
        :param bank_code: 银行编号
        :param amount: 付款金额，单位分
        :param desc: 付款说明
        :param out_trade_no: 可选，商户订单号，需保持唯一性，默认自动生成
        :return: 返回的结果信息
        """
        if not out_trade_no:
            now = datetime.now()
            out_trade_no = '{0}{1}{2}'.format(
                self.mch_id,
                now.strftime('%Y%m%d%H%M%S'),
                random.randint(1000, 10000)
            )
        data = {
            'mch_id': self.mch_id,
            'partner_trade_no': out_trade_no,
            'amount': amount,
            'desc': desc,
            'enc_bank_no': self._rsa_encrypt(bank_card_no),
            'enc_true_name': self._rsa_encrypt(true_name),
            'bank_code': bank_code,
        }
        return self._post('mmpaysptrans/pay_bank', data=data)

    def query_bankcard(self, out_trade_no):
        """
        企业付款查询接口

        :param out_trade_no: 商户调用企业付款API时使用的商户订单号
        :return: 返回的结果数据
        """
        data = {
            'mch_id': self.mch_id,
            'partner_trade_no': out_trade_no,
        }
        return self._post('mmpaysptrans/query_bank', data=data)

    def get_rsa_public_key(self):
        data = {
            'mch_id': self.mch_id,
            'sign_type': 'MD5',
        }
        return self._post('https://fraud.mch.weixin.qq.com/risk/getpublickey', data=data)

    def _rsa_encrypt(self, data):
        if not getattr(self, '_rsa_public_key', None):
            self._rsa_public_key = self.get_rsa_public_key()['pub_key']
        return rsa_encrypt(data, self._rsa_public_key)



# -*- coding: utf-8 -*-
"""
    wechatpyyyy._compat
    ~~~~~~~~~~~~~~~~~

    This module makes it easy for wechatpy to run on both Python 2 and 3.

    :copyright: (c) 2014 by messense.
    :license: MIT, see LICENSE for more details.
"""
from __future__ import absolute_import, unicode_literals
import sys  # NOQA
import six  # NOQA
import warnings

warnings.warn("Module `wechatpyyyy._compat` is deprecated, will be removed in 2.0"
              "use `wechatpyyyy.utils` instead",
              DeprecationWarning, stacklevel=2)

from wechatpyyyy.utils import get_querystring  # NOQA
from wechatpyyyy.utils import json  # NOQA



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from wechatpyyyy.client.api.base import BaseWeChatAPI
from wechatpyyyy.utils import json


class WeChatMaterial(BaseWeChatAPI):

    def add_articles(self, articles):
        """
        新增永久图文素材
        详情请参考
        https://developers.weixin.qq.com/doc/offiaccount/Asset_Management/Adding_Permanent_Assets.html

        :param articles: 图文素材数组
        :type articles: list[dict]
        :return: 返回的 JSON 数据包
        """
        articles_data = []
        for article in articles:
            articles_data.append({
                'thumb_media_id': article['thumb_media_id'],
                'title': article['title'],
                'content': article['content'],
                'author': article.get('author', ''),
                'content_source_url': article.get('content_source_url', ''),
                'digest': article.get('digest', ''),
                'show_cover_pic': article.get('show_cover_pic', 0),
                'need_open_comment': int(article.get('need_open_comment', False)),
                'only_fans_can_comment': int(article.get('only_fans_can_comment', False)),
            })
        return self._post(
            'material/add_news',
            data={
                'articles': articles_data
            }
        )

    def add(self, media_type, media_file, title=None, introduction=None):
        """
        新增其它类型永久素材
        详情请参考
        https://developers.weixin.qq.com/doc/offiaccount/Asset_Management/Adding_Permanent_Assets.html

        :param media_type: 媒体文件类型，分别有图片（image）、语音（voice）、视频（video）和缩略图（thumb）
        :param media_file: 要上传的文件，一个 File-object
        :param title: 视频素材标题，仅上传视频素材时需要
        :param introduction: 视频素材简介，仅上传视频素材时需要
        :return: 返回的 JSON 数据包
        """
        params = {
            'access_token': self.access_token,
            'type': media_type
        }
        if media_type == 'video':
            assert title, 'Video title must be set'
            assert introduction, 'Video introduction must be set'
            description = {
                'title': title,
                'introduction': introduction
            }
            params['description'] = json.dumps(description)
        return self._post(
            'material/add_material',
            params=params,
            files={
                'media': media_file
            }
        )

    def get(self, media_id):
        """
        获取永久素材
        详情请参考
        https://developers.weixin.qq.com/doc/offiaccount/Asset_Management/Getting_Permanent_Assets.html

        :param media_id: 素材的 media_id
        :return: 图文素材返回图文列表，其它类型为素材的内容
        """

        def _processor(res):
            if isinstance(res, dict):
                if 'news_item' in res:
                    # 图文素材
                    return res['news_item']
            return res

        return self._post(
            'material/get_material',
            data={
                'media_id': media_id
            },
            result_processor=_processor
        )

    def delete(self, media_id):
        """
        删除永久素材
        详情请参考
        https://developers.weixin.qq.com/doc/offiaccount/Asset_Management/Deleting_Permanent_Assets.html

        :param media_id: 素材的 media_id
        :return: 返回的 JSON 数据包
        """
        return self._post(
            'material/del_material',
            data={
                'media_id': media_id
            }
        )

    def update_article(self, media_id, index, article):
        """
        修改永久图文素材
        详情请参考
        https://developers.weixin.qq.com/doc/offiaccount/Asset_Management/Editing_Permanent_Rich_Media_Assets.html

        :param media_id: 要修改的图文消息的 id
        :param index: 要更新的文章在图文消息中的位置（多图文消息时，此字段才有意义），第一篇为 0
        :param article: 图文素材
        :return: 返回的 JSON 数据包
        """
        article_data = {
            'thumb_media_id': article['thumb_media_id'],
            'title': article['title'],
            'content': article['content'],
            'author': article.get('author', ''),
            'content_source_url': article.get('content_source_url', ''),
            'digest': article.get('digest', ''),
            'show_cover_pic': article.get('show_cover_pic', 0)
        }
        return self._post(
            'material/update_news',
            data={
                'media_id': media_id,
                'index': index,
                'articles': article_data
            }
        )

    def update_articles(self, media_id, index, articles):
        """
        修改永久图文素材
        详情请参考
        https://developers.weixin.qq.com/doc/offiaccount/Asset_Management/Editing_Permanent_Rich_Media_Assets.html

        :param media_id: 要修改的图文消息的 id
        :param index: 要更新的文章在图文消息中的位置（多图文消息时，此字段才有意义），第一篇为 0
        :param articles: 图文素材数组
        :return: 返回的 JSON 数据包
        """
        return self.update_article(media_id, index, articles[index])

    def batchget(self, media_type, offset=0, count=20):
        """
        获取素材列表
        详情请参考
        https://developers.weixin.qq.com/doc/offiaccount/Asset_Management/Get_materials_list.html

        :param media_type: 媒体文件类型，分别有图片（image）、语音（voice）、视频（video）和缩略图（news）
        :param offset: 从全部素材的该偏移位置开始返回，0 表示从第一个素材返回
        :param count: 返回素材的数量，取值在1到20之间
        :return: 返回的 JSON 数据包
        """
        return self._post(
            'material/batchget_material',
            data={
                'type': media_type,
                'offset': offset,
                'count': count
            }
        )

    def get_count(self):
        """
        获取素材总数
        详情请参考
        https://developers.weixin.qq.com/doc/offiaccount/Asset_Management/Get_the_total_of_all_materials.html

        :return: 返回的 JSON 数据包
        """
        return self._get('material/get_materialcount')

    def open_comment(self, msg_data_id, index=1):
        """
        打开已群发文章评论
        https://mp.weixin.qq.com/wiki?id=mp1494572718_WzHIY
        """
        return self._post(
            'comment/open',
            data={
                'msg_data_id': msg_data_id,
                'index': index,
            })

    def close_comment(self, msg_data_id, index=1):
        """
        关闭已群发文章评论
        """
        return self._post(
            'comment/close',
            data={
                'msg_data_id': msg_data_id,
                'index': index,
            })

    def list_comment(self, msg_data_id, index=1, begin=0, count=50, type=0):
        """
        查看指定文章的评论数据
        """
        return self._post(
            'comment/list',
            data={
                'msg_data_id': msg_data_id,
                'index': index,
                'begin': begin,
                'count': count,
                'type': type
            })

    def markelect_comment(self, msg_data_id, index, user_comment_id):
        """
        将评论标记精选
        """
        return self._post(
            'comment/markelect',
            data={
                'msg_data_id': msg_data_id,
                'index': index,
                'user_comment_id': user_comment_id,
            })

    def unmarkelect_comment(self, msg_data_id, index, user_comment_id):
        """
        将评论取消精选
        """
        return self._post(
            'comment/unmarkelect',
            data={
                'msg_data_id': msg_data_id,
                'index': index,
                'user_comment_id': user_comment_id,
            })

    def delete_comment(self, msg_data_id, index, user_comment_id):
        """
        删除评论
        """
        return self._post(
            'comment/delete',
            data={
                'msg_data_id': msg_data_id,
                'index': index,
                'user_comment_id': user_comment_id,
            })

    def add_reply_comment(self, msg_data_id, index, user_comment_id, content):
        """
        回复评论
        """
        return self._post(
            'comment/reply/add',
            data={
                'msg_data_id': msg_data_id,
                'index': index,
                'user_comment_id': user_comment_id,
                'content': content
            })

    def delete_reply_comment(self, msg_data_id, index, user_comment_id):
        """
        删除回复
        """
        return self._post(
            'comment/reply/delete',
            data={
                'msg_data_id': msg_data_id,
                'index': index,
                'user_comment_id': user_comment_id,
            })



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from wechatpyyyy.client.api.base import BaseWeChatAPI


class WeChatTag(BaseWeChatAPI):
    """
    标签管理

    https://work.weixin.qq.com/api/doc#90000/90135/90209
    """

    def create(self, name):
        return self._post(
            'tag/create',
            data={
                'tagname': name
            }
        )

    def update(self, tag_id, name):
        return self._post(
            'tag/update',
            data={
                'tagid': tag_id,
                'tagname': name
            }
        )

    def delete(self, tag_id):
        return self._get(
            'tag/delete',
            params={
                'tagid': tag_id
            }
        )

    def get_users(self, tag_id):
        return self._get(
            'tag/get',
            params={
                'tagid': tag_id
            }
        )

    def add_users(self, tag_id, user_ids):
        return self._post(
            'tag/addtagusers',
            data={
                'tagid': tag_id,
                'userlist': user_ids
            }
        )

    def delete_users(self, tag_id, user_ids):
        return self._post(
            'tag/deltagusers',
            data={
                'tagid': tag_id,
                'userlist': user_ids
            }
        )

    def list(self):
        res = self._get('tag/list')
        return res['taglist']



# -*- coding: utf-8 -*-
"""
    wechatpyyyy.messages
    ~~~~~~~~~~~~~~~~~~

    This module defines all the messages you can get from WeChat server

    :copyright: (c) 2014 by messense.
    :license: MIT, see LICENSE for more details.
"""
from __future__ import absolute_import, unicode_literals
import copy
import six

from wechatpyyyy.fields import (
    BaseField,
    StringField,
    IntegerField,
    DateTimeField,
    FieldDescriptor
)
from wechatpyyyy.utils import to_text, to_binary


MESSAGE_TYPES = {}


def register_message(msg_type):
    def register(cls):
        MESSAGE_TYPES[msg_type] = cls
        return cls
    return register


class MessageMetaClass(type):
    """Metaclass for all messages"""
    def __new__(cls, name, bases, attrs):
        for b in bases:
            if not hasattr(b, '_fields'):
                continue

            for k, v in b.__dict__.items():
                if k in attrs:
                    continue
                if isinstance(v, FieldDescriptor):
                    attrs[k] = copy.deepcopy(v.field)

        cls = super(MessageMetaClass, cls).__new__(cls, name, bases, attrs)
        cls._fields = {}

        for name, field in cls.__dict__.items():
            if isinstance(field, BaseField):
                field.add_to_class(cls, name)
        return cls


class BaseMessage(six.with_metaclass(MessageMetaClass)):
    """Base class for all messages and events"""
    type = 'unknown'
    id = IntegerField('MsgId', 0)
    source = StringField('FromUserName')
    target = StringField('ToUserName')
    create_time = DateTimeField('CreateTime')
    time = IntegerField('CreateTime')

    def __init__(self, message):
        self._data = message

    def __repr__(self):
        _repr = "{klass}({msg})".format(
            klass=self.__class__.__name__,
            msg=repr(self._data)
        )
        if six.PY2:
            return to_binary(_repr)
        else:
            return to_text(_repr)


@register_message('text')
class TextMessage(BaseMessage):
    """
    文本消息
    详情请参阅
    http://mp.weixin.qq.com/wiki/10/79502792eef98d6e0c6e1739da387346.html
    """
    type = 'text'
    content = StringField('Content')


@register_message('image')
class ImageMessage(BaseMessage):
    """
    图片消息
    详情请参阅
    http://mp.weixin.qq.com/wiki/10/79502792eef98d6e0c6e1739da387346.html
    """
    type = 'image'
    media_id = StringField('MediaId')
    image = StringField('PicUrl')


@register_message('voice')
class VoiceMessage(BaseMessage):
    """
    语音消息
    详情请参阅
    http://mp.weixin.qq.com/wiki/10/79502792eef98d6e0c6e1739da387346.html
    """
    type = 'voice'
    media_id = StringField('MediaId')
    format = StringField('Format')
    recognition = StringField('Recognition')


@register_message('shortvideo')
class ShortVideoMessage(BaseMessage):
    """
    短视频消息
    详情请参阅
    http://mp.weixin.qq.com/wiki/10/79502792eef98d6e0c6e1739da387346.html
    """
    type = 'shortvideo'
    media_id = StringField('MediaId')
    thumb_media_id = StringField('ThumbMediaId')


@register_message('video')
class VideoMessage(BaseMessage):
    """
    视频消息
    详情请参阅
    http://mp.weixin.qq.com/wiki/10/79502792eef98d6e0c6e1739da387346.html
    """
    type = 'video'
    media_id = StringField('MediaId')
    thumb_media_id = StringField('ThumbMediaId')


@register_message('location')
class LocationMessage(BaseMessage):
    """
    地理位置消息
    详情请参阅
    http://mp.weixin.qq.com/wiki/10/79502792eef98d6e0c6e1739da387346.html
    """
    type = 'location'
    location_x = StringField('Location_X')
    location_y = StringField('Location_Y')
    scale = StringField('Scale')
    label = StringField('Label')

    @property
    def location(self):
        return self.location_x, self.location_y


@register_message('link')
class LinkMessage(BaseMessage):
    """
    链接消息
    详情请参阅
    http://mp.weixin.qq.com/wiki/10/79502792eef98d6e0c6e1739da387346.html
    """
    type = 'link'
    title = StringField('Title')
    description = StringField('Description')
    url = StringField('Url')


@register_message('miniprogrampage')
class MiniProgramPageMessage(BaseMessage):
    """
    小程序卡片消息
    详情请参阅
    https://developers.weixin.qq.com/miniprogram/dev/framework/open-ability/customer-message/receive.html#小程序卡片消息
    """
    type = 'miniprogrampage'
    app_id = StringField('AppId')
    title = StringField('Title')
    page_path = StringField('PagePath')
    thumb_url = StringField('ThumbUrl')
    thumb_media_id = StringField('ThumbMediaId')


class UnknownMessage(BaseMessage):
    """未知消息类型"""
    pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from wechatpyyyy.utils import to_text
from wechatpyyyy.client.api.base import BaseWeChatAPI


class WeChatTag(BaseWeChatAPI):

    def create(self, name):
        """
        创建标签

        :param name: 标签名（30个字符以内）
        :return: 返回的 JSON 数据包

        """
        name = to_text(name)
        return self._post(
            'tags/create',
            data={'tag': {'name': name}},
            result_processor=lambda x: x['tag']
        )

    def get(self):
        """
        获取公众号已创建的标签

        :return: 所有标签列表

        """

        res = self._get(
            'tags/get',
            result_processor=lambda x: x['tags']
        )

        return res

    def update(self, tag_id, name):
        """
        编辑标签

        :param tag_id: 标签id，由微信分配
        :param name: 标签名字（30个字符以内）
        :return: 返回的 JSON 数据包

        """
        name = to_text(name)
        return self._post(
            'tags/update',
            data={
                'tag': {
                    'id': int(tag_id),
                    'name': name
                }
            }
        )

    def delete(self, tag_id):
        """
        删除标签

        :param tag_id: 标签 ID
        :return: 返回的 JSON 数据包

        """
        return self._post(
            'tags/delete',
            data={
                'tag': {
                    'id': tag_id
                }
            }
        )

    def tag_user(self, tag_id, user_id):
        """
        批量为用户打标签

        :param tag_id: 标签 ID
        :param user_id: 用户 ID, 可以是单个或者列表

        :return: 返回的 JSON 数据包

        """
        data = {'tagid': tag_id}
        if isinstance(user_id, (tuple, list)):
            data['openid_list'] = user_id
        else:
            data['openid_list'] = [user_id, ]
        return self._post('tags/members/batchtagging', data=data)

    def untag_user(self, tag_id, user_id):
        """
        批量为用户取消标签

        :param tag_id: 标签 ID
        :param user_id: 用户 ID, 可以是单个或者列表

        :return: 返回的 JSON 数据包

        """
        data = {'tagid': tag_id}
        if isinstance(user_id, (tuple, list)):
            data['openid_list'] = user_id
        else:
            data['openid_list'] = [user_id, ]
        return self._post('tags/members/batchuntagging', data=data)

    def get_user_tag(self, user_id):
        """
        获取用户身上的标签列表

        :param user_id: 用户 ID, 可以是单个或者列表
        :return: 返回的 JSON 数据包
        """
        return self._post(
            'tags/getidlist',
            data={
                'openid': user_id
            },
            result_processor=lambda x: x['tagid_list']
        )

    def get_tag_users(self, tag_id, first_user_id=None):
        """
        获取标签下粉丝列表

        :param tag_id: 标签 ID
        :param first_user_id: 可选。第一个拉取的 OPENID，不填默认从头开始拉取
        :return: 返回的 JSON 数据包
        """
        data = {
            'tagid': tag_id,
        }
        if first_user_id:
            data['next_openid'] = first_user_id
        return self._post(
            'user/tag/get',
            data=data
        )

    def iter_tag_users(self, tag_id, first_user_id=None):
        """
        获取标签下粉丝openid列表

        :return: 返回一个迭代器，可以用for进行循环，得到openid

        使用示例::

            from wechatpy import WeChatClient

            client = WeChatClient('appid', 'secret')
            for openid in client.tag.iter_tag_users(0):
                print(openid)

        """
        while True:
            follower_data = self.get_tag_users(tag_id, first_user_id)
            if 'data' not in follower_data:
                return
            for openid in follower_data['data']['openid']:
                yield openid
            first_user_id = follower_data.get('next_openid')
            if not first_user_id:
                return

    def get_black_list(self, begin_openid=None):
        """
        获取公众号的黑名单列表
        详情请参考
        https://mp.weixin.qq.com/wiki?id=mp1471422259_pJMWA

        :param begin_openid: 起始的 OpenID，传空则默认从头开始拉取
        :return: 返回的 JSON 数据包
        :rtype: dict
        """
        data = {}
        if begin_openid:
            data['begin_openid'] = begin_openid
        return self._post(
            'tags/members/getblacklist',
            data=data,
        )

    def batch_black_list(self, openid_list):
        """
        批量拉黑用户
        详情请参考
        https://mp.weixin.qq.com/wiki?id=mp1471422259_pJMWA

        :param openid_list: 批量拉黑用户的 OpenID list, 最多20个
        :type openid_list: list
        """
        return self._post(
            'tags/members/batchblacklist',
            data={
                'openid_list': openid_list,
            },
        )

    def batch_unblack_list(self, openid_list):
        """
        批量取消拉黑
        详情请参考
        https://mp.weixin.qq.com/wiki?id=mp1471422259_pJMWA

        :param openid_list: 批量取消拉黑的 OpenID list, 最多20个
        :type openid_list: list
        """
        return self._post(
            'tags/members/batchunblacklist',
            data={
                'openid_list': openid_list,
            },
        )



# encoding: utf-8
from __future__ import absolute_import, unicode_literals

import six

from wechatpyyyy.client.api.base import BaseWeChatAPI


class WeChatOAuth(BaseWeChatAPI):
    OAUTH_BASE_URL = 'https://open.weixin.qq.com/connect/oauth2/authorize'

    def authorize_url(self, redirect_uri, state=None):
        """
        构造网页授权链接
        详情请参考
        https://work.weixin.qq.com/api/doc#90000/90135/91022

        :param redirect_uri: 授权后重定向的回调链接地址
        :param state: 重定向后会带上 state 参数
        :return: 返回的 JSON 数据包
        """
        redirect_uri = six.moves.urllib.parse.quote(redirect_uri, safe=b'')
        url_list = [
            self.OAUTH_BASE_URL,
            '?appid=',
            self._client.corp_id,
            '&redirect_uri=',
            redirect_uri,
            '&response_type=code&scope=snsapi_base',
        ]
        if state:
            url_list.extend(['&state=', state])
        url_list.append('#wechat_redirect')
        return ''.join(url_list)

    def get_user_info(self, code):
        """
        获取访问用户身份
        详情请参考
        https://work.weixin.qq.com/api/doc#90000/90135/91023

        :param code: 通过成员授权获取到的code
        :return: 返回的 JSON 数据包
        """

        return self._get(
            'user/getuserinfo',
            params={
                'code': code,
            }
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from optionaldict import optionaldict

from wechatpyyyy.client.api.base import BaseWeChatAPI


class WeChatChat(BaseWeChatAPI):
    """
    “微信企业号”旧接口，企业微信请使用 appchat。
    """

    def create(self, chat_id, name, owner, user_list):
        """
        创建会话

        详情请参考
        https://qydev.weixin.qq.com/wiki/index.php?title=企业会话接口说明

        :param chat_id: 会话id。字符串类型，最长32个字符。只允许字符0-9及字母a-zA-Z,
                        如果值内容为64bit无符号整型：要求值范围在[1, 2^63)之间，
                        [2^63, 2^64)为系统分配会话id区间
        :param name: 会话标题
        :param owner: 管理员userid，必须是该会话userlist的成员之一
        :param user_list: 会话成员列表，成员用userid来标识。会话成员必须在3人或以上，1000人以下
        :return: 返回的 JSON 数据包
        """
        return self._post(
            'chat/create',
            data={
                'chatid': chat_id,
                'name': name,
                'owner': owner,
                'userlist': user_list,
            }
        )

    def get(self, chat_id):
        """
        获取会话

        详情请参考
        https://qydev.weixin.qq.com/wiki/index.php?title=企业会话接口说明

        :param chat_id: 会话 ID
        :return: 会话信息
        """
        res = self._get('chat/get', params={'chatid': chat_id})
        return res['chat_info']

    def update(self, chat_id, op_user, name=None, owner=None,
               add_user_list=None, del_user_list=None):
        """
        修改会话

        详情请参考
        https://qydev.weixin.qq.com/wiki/index.php?title=企业会话接口说明

        :param chat_id: 会话 ID
        :param op_user: 操作人 userid
        :param name: 会话标题
        :param owner: 管理员userid，必须是该会话userlist的成员之一
        :param add_user_list: 会话新增成员列表，成员用userid来标识
        :param del_user_list: 会话退出成员列表，成员用userid来标识
        :return: 返回的 JSON 数据包
        """
        data = optionaldict(
            chatid=chat_id,
            op_user=op_user,
            name=name,
            owner=owner,
            add_user_list=add_user_list,
            del_user_list=del_user_list,
        )
        return self._post('chat/update', data=data)

    def quit(self, chat_id, op_user):
        """
        退出会话

        详情请参考
        https://qydev.weixin.qq.com/wiki/index.php?title=企业会话接口说明

        :param chat_id: 会话 ID
        :param op_user: 操作人 userid
        :return: 返回的 JSON 数据包
        """
        return self._post(
            'chat/quit',
            data={
                'chatid': chat_id,
                'op_user': op_user,
            }
        )

    def clear_notify(self, op_user, type, id):
        """
        清除会话未读状态

        详情请参考
        https://qydev.weixin.qq.com/wiki/index.php?title=企业会话接口说明

        :param op_user: 会话所有者的userid
        :param type: 会话类型：single|group，分别表示：单聊|群聊
        :param id: 会话值，为userid|chatid，分别表示：成员id|会话id
        :return: 返回的 JSON 数据包
        """
        return self._post(
            'chat/clearnotify',
            data={
                'op_user': op_user,
                'chat': {
                    'type': type,
                    'id': id,
                }
            }
        )

    def set_mute(self, user_mute_list):
        """
        设置成员新消息免打扰

        详情请参考
        https://qydev.weixin.qq.com/wiki/index.php?title=企业会话接口说明

        :param user_mute_list: 成员新消息免打扰参数，数组，最大支持10000个成员
        :return: 返回的 JSON 数据包
        """
        return self._post(
            'chat/setmute',
            data={'user_mute_list': user_mute_list}
        )

    def send_text(self, sender, receiver_type, receiver_id, content):
        """
        发送文本消息

        详情请参考
        https://qydev.weixin.qq.com/wiki/index.php?title=企业会话接口说明

        :param sender: 发送人
        :param receiver_type: 接收人类型：single|group，分别表示：单聊|群聊
        :param receiver_id: 接收人的值，为userid|chatid，分别表示：成员id|会话id
        :param content: 消息内容
        :return: 返回的 JSON 数据包
        """
        data = {
            'receiver': {
                'type': receiver_type,
                'id': receiver_id,
            },
            'sender': sender,
            'msgtype': 'text',
            'text': {
                'content': content,
            }
        }
        return self._post('chat/send', data=data)

    def send_single_text(self, sender, receiver, content):
        """
        发送单聊文本消息

        :param sender: 发送人
        :param receiver: 接收人成员 ID
        :param content: 消息内容
        :return: 返回的 JSON 数据包
        """
        return self.send_text(sender, 'single', receiver, content)

    def send_group_text(self, sender, receiver, content):
        """
        发送群聊文本消息

        :param sender: 发送人
        :param receiver: 会话 ID
        :param content: 消息内容
        :return: 返回的 JSON 数据包
        """
        return self.send_text(sender, 'group', receiver, content)

    def send_image(self, sender, receiver_type, receiver_id, media_id):
        """
        发送图片消息

        详情请参考
        https://qydev.weixin.qq.com/wiki/index.php?title=企业会话接口说明

        :param sender: 发送人
        :param receiver_type: 接收人类型：single|group，分别表示：单聊|群聊
        :param receiver_id: 接收人的值，为userid|chatid，分别表示：成员id|会话id
        :param media_id: 图片媒体文件id，可以调用上传素材文件接口获取
        :return: 返回的 JSON 数据包
        """
        data = {
            'receiver': {
                'type': receiver_type,
                'id': receiver_id,
            },
            'sender': sender,
            'msgtype': 'image',
            'image': {
                'media_id': media_id,
            }
        }
        return self._post('chat/send', data=data)

    def send_single_image(self, sender, receiver, media_id):
        """
        发送单聊图片消息

        :param sender: 发送人
        :param receiver: 接收人成员 ID
        :param media_id: 图片媒体文件id，可以调用上传素材文件接口获取
        :return: 返回的 JSON 数据包
        """
        return self.send_image(sender, 'single', receiver, media_id)

    def send_group_image(self, sender, receiver, media_id):
        """
        发送群聊图片消息

        :param sender: 发送人
        :param receiver: 会话 ID
        :param media_id: 图片媒体文件id，可以调用上传素材文件接口获取
        :return: 返回的 JSON 数据包
        """
        return self.send_image(sender, 'group', receiver, media_id)

    def send_file(self, sender, receiver_type, receiver_id, media_id):
        """
        发送文件消息

        详情请参考
        https://qydev.weixin.qq.com/wiki/index.php?title=企业会话接口说明

        :param sender: 发送人
        :param receiver_type: 接收人类型：single|group，分别表示：单聊|群聊
        :param receiver_id: 接收人的值，为userid|chatid，分别表示：成员id|会话id
        :param media_id: 文件id，可以调用上传素材文件接口获取, 文件须大于4字节
        :return: 返回的 JSON 数据包
        """
        data = {
            'receiver': {
                'type': receiver_type,
                'id': receiver_id,
            },
            'sender': sender,
            'msgtype': 'file',
            'file': {
                'media_id': media_id,
            }
        }
        return self._post('chat/send', data=data)

    def send_single_file(self, sender, receiver, media_id):
        """
        发送单聊文件消息

        :param sender: 发送人
        :param receiver: 接收人成员 ID
        :param media_id: 文件id，可以调用上传素材文件接口获取, 文件须大于4字节
        :return: 返回的 JSON 数据包
        """
        return self.send_file(sender, 'single', receiver, media_id)

    def send_group_file(self, sender, receiver, media_id):
        """
        发送群聊文件消息

        :param sender: 发送人
        :param receiver: 会话 ID
        :param media_id: 文件id，可以调用上传素材文件接口获取, 文件须大于4字节
        :return: 返回的 JSON 数据包
        """
        return self.send_file(sender, 'group', receiver, media_id)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import six

from wechatpy import replies
from wechatpyyyy.fields import IntegerField

REPLY_TYPES = {}


def register_reply(reply_type):
    def register(cls):
        REPLY_TYPES[reply_type] = cls
        return cls

    return register


@register_reply('text')
class TextReply(replies.TextReply):
    agent = IntegerField('AgentID', 0)


@register_reply('image')
class ImageReply(replies.ImageReply):
    agent = IntegerField('AgentID', 0)


@register_reply('voice')
class VoiceReply(replies.VoiceReply):
    agent = IntegerField('AgentID', 0)


@register_reply('video')
class VideoReply(replies.VideoReply):
    agent = IntegerField('AgentID', 0)


@register_reply('news')
class ArticlesReply(replies.ArticlesReply):
    agent = IntegerField('AgentID', 0)


def create_reply(reply, message=None, render=False):
    r = None
    if isinstance(reply, replies.BaseReply):
        r = reply
        if message:
            r.source = message.target
            r.target = message.source
            r.agent = message.agent
    elif isinstance(reply, six.string_types):
        r = TextReply(
            message=message,
            content=reply
        )
    elif isinstance(reply, (tuple, list)):
        if len(reply) > 10:
            raise AttributeError("Can't add more than 10 articles"
                                 " in an ArticlesReply")
        r = ArticlesReply(
            message=message,
            articles=reply
        )
    if r and render:
        return r.render()
    return r



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
from wechatpyyyy.client.api.base import BaseWeChatAPI


class MerchantStock(BaseWeChatAPI):

    API_BASE_URL = 'https://api.weixin.qq.com/'

    def add(self, product_id, quantity, sku_info=''):
        return self._post(
            'merchant/stock/add',
            data={
                'product_id': product_id,
                'quantity': quantity,
                'sku_info': sku_info
            }
        )

    def reduce(self, product_id, quantity, sku_info=''):
        return self._post(
            'merchant/stock/reduce',
            data={
                'product_id': product_id,
                'quantity': quantity,
                'sku_info': sku_info
            }
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from wechatpy import events
from wechatpyyyy.fields import BaseField, IntegerField

EVENT_TYPES = {}


def register_event(event_type):
    def register(cls):
        EVENT_TYPES[event_type] = cls
        return cls

    return register


@register_event('subscribe')
class SubscribeEvent(events.SubscribeEvent):
    """
    成员关注事件
    详情请参阅
    https://qydev.weixin.qq.com/wiki/index.php?title=接收事件#.E6.88.90.E5.91.98.E5.85.B3.E6.B3.A8.2F.E5.8F.96.E6.B6.88.E5.85.B3.E6.B3.A8.E4.BA.8B.E4.BB.B6
    """
    agent = IntegerField('AgentID', 0)
    event = 'subscribe'


@register_event('unsubscribe')
class UnsubscribeEvent(events.UnsubscribeEvent):
    """
    成员取消关注事件
    详情请参阅
    https://qydev.weixin.qq.com/wiki/index.php?title=接收事件#.E6.88.90.E5.91.98.E5.85.B3.E6.B3.A8.2F.E5.8F.96.E6.B6.88.E5.85.B3.E6.B3.A8.E4.BA.8B.E4.BB.B6
    """
    agent = IntegerField('AgentID', 0)
    event = 'unsubscribe'


@register_event('click')
class ClickEvent(events.ClickEvent):
    """
    点击菜单拉取消息事件
    详情请参阅
    http://qydev.weixin.qq.com/wiki/index.php?title=接收事件#.E7.82.B9.E5.87.BB.E8.8F.9C.E5.8D.95.E6.8B.89.E5.8F.96.E6.B6.88.E6.81.AF.E7.9A.84.E4.BA.8B.E4.BB.B6.E6.8E.A8.E9.80.81
    """
    agent = IntegerField('AgentID', 0)
    event = 'click'


@register_event('view')
class ViewEvent(events.ViewEvent):
    """
    点击菜单跳转链接事件
    详情请参阅
    http://qydev.weixin.qq.com/wiki/index.php?title=接收事件#.E7.82.B9.E5.87.BB.E8.8F.9C.E5.8D.95.E8.B7.B3.E8.BD.AC.E9.93.BE.E6.8E.A5.E7.9A.84.E4.BA.8B.E4.BB.B6.E6.8E.A8.E9.80.81
    """
    agent = IntegerField('AgentID', 0)
    event = 'view'


@register_event('location')
class LocationEvent(events.LocationEvent):
    """
    上报地理位置事件
    详情请参阅
    http://qydev.weixin.qq.com/wiki/index.php?title=接收事件#.E4.B8.8A.E6.8A.A5.E5.9C.B0.E7.90.86.E4.BD.8D.E7.BD.AE.E4.BA.8B.E4.BB.B6
    """
    agent = IntegerField('AgentID', 0)
    event = 'location'


@register_event('scancode_push')
class ScanCodePushEvent(events.ScanCodePushEvent):
    """
    扫码推事件的事件
    详情请参阅
    http://qydev.weixin.qq.com/wiki/index.php?title=接收事件#.E6.89.AB.E7.A0.81.E6.8E.A8.E4.BA.8B.E4.BB.B6.E7.9A.84.E4.BA.8B.E4.BB.B6.E6.8E.A8.E9.80.81
    """
    agent = IntegerField('AgentID', 0)
    event = 'scancode_push'


@register_event('scancode_waitmsg')
class ScanCodeWaitMsgEvent(events.ScanCodeWaitMsgEvent):
    """
    扫码推事件且弹出“消息接收中”提示框的事件
    详情请参阅
    http://qydev.weixin.qq.com/wiki/index.php?title=接收事件#.E6.89.AB.E7.A0.81.E6.8E.A8.E4.BA.8B.E4.BB.B6.E4.B8.94.E5.BC.B9.E5.87.BA.E2.80.9C.E6.B6.88.E6.81.AF.E6.8E.A5.E6.94.B6.E4.B8.AD.E2.80.9D.E6.8F.90.E7.A4.BA.E6.A1.86.E7.9A.84.E4.BA.8B.E4.BB.B6.E6.8E.A8.E9.80.81
    """
    agent = IntegerField('AgentID', 0)
    event = 'scancode_waitmsg'


@register_event('pic_sysphoto')
class PicSysPhotoEvent(events.PicSysPhotoEvent):
    """
    弹出系统拍照发图事件
    详情请参阅
    http://qydev.weixin.qq.com/wiki/index.php?title=接收事件#.E5.BC.B9.E5.87.BA.E7.B3.BB.E7.BB.9F.E6.8B.8D.E7.85.A7.E5.8F.91.E5.9B.BE.E7.9A.84.E4.BA.8B.E4.BB.B6.E6.8E.A8.E9.80.81
    """
    agent = IntegerField('AgentID', 0)
    event = 'pic_sysphoto'


@register_event('pic_photo_or_album')
class PicPhotoOrAlbumEvent(events.PicPhotoOrAlbumEvent):
    """
    弹出拍照或相册发图事件
    详情请参阅
    http://qydev.weixin.qq.com/wiki/index.php?title=接收事件#.E5.BC.B9.E5.87.BA.E6.8B.8D.E7.85.A7.E6.88.96.E8.80.85.E7.9B.B8.E5.86.8C.E5.8F.91.E5.9B.BE.E7.9A.84.E4.BA.8B.E4.BB.B6.E6.8E.A8.E9.80.81
    """
    agent = IntegerField('AgentID', 0)
    event = 'pic_photo_or_album'


@register_event('pic_weixin')
class PicWeChatEvent(events.PicWeChatEvent):
    """
    弹出微信相册发图器事件
    详情请参阅
    http://qydev.weixin.qq.com/wiki/index.php?title=接收事件#.E5.BC.B9.E5.87.BA.E5.BE.AE.E4.BF.A1.E7.9B.B8.E5.86.8C.E5.8F.91.E5.9B.BE.E5.99.A8.E7.9A.84.E4.BA.8B.E4.BB.B6.E6.8E.A8.E9.80.81
    """
    agent = IntegerField('AgentID', 0)
    event = 'pic_weixin'


@register_event('location_select')
class LocationSelectEvent(events.LocationSelectEvent):
    """
    弹出地理位置选择器事件
    详情请参阅
    http://qydev.weixin.qq.com/wiki/index.php?title=接收事件#.E5.BC.B9.E5.87.BA.E5.9C.B0.E7.90.86.E4.BD.8D.E7.BD.AE.E9.80.89.E6.8B.A9.E5.99.A8.E7.9A.84.E4.BA.8B.E4.BB.B6.E6.8E.A8.E9.80.81
    """
    agent = IntegerField('AgentID', 0)
    event = 'location_select'


@register_event('enter_agent')
class EnterAgentEvent(events.BaseEvent):
    """
    用户进入应用的事件推送
    详情请参阅
    http://qydev.weixin.qq.com/wiki/index.php?title=接收事件#.E6.88.90.E5.91.98.E8.BF.9B.E5.85.A5.E5.BA.94.E7.94.A8.E7.9A.84.E4.BA.8B.E4.BB.B6.E6.8E.A8.E9.80.81
    """
    agent = IntegerField('AgentID', 0)
    event = 'enter_agent'


@register_event('batch_job_result')
class BatchJobResultEvent(events.BaseEvent):
    """
    异步任务完成事件
    详情请参阅
    http://qydev.weixin.qq.com/wiki/index.php?title=接收事件#.E5.BC.82.E6.AD.A5.E4.BB.BB.E5.8A.A1.E5.AE.8C.E6.88.90.E4.BA.8B.E4.BB.B6.E6.8E.A8.E9.80.81
    """
    event = 'batch_job_result'
    batch_job = BaseField('BatchJob')

    @property
    def job_id(self):
        return self.batch_job['JobId']

    @property
    def job_type(self):
        return self.batch_job['JobType']

    @property
    def err_code(self):
        return self.batch_job['ErrCode']

    @property
    def err_msg(self):
        return self.batch_job['ErrMsg']



from __future__ import absolute_import, unicode_literals

from wechatpyyyy.crypto import BasePrpCrypto, BaseWeChatCrypto
from wechatpyyyy.enterprise.exceptions import InvalidCorpIdException


class PrpCrypto(BasePrpCrypto):

    def encrypt(self, text, corp_id):
        return self._encrypt(text, corp_id)

    def decrypt(self, text, corp_id):
        return self._decrypt(text, corp_id, InvalidCorpIdException)


class WeChatCrypto(BaseWeChatCrypto):

    def __init__(self, token, encoding_aes_key, corp_id):
        super(WeChatCrypto, self).__init__(token, encoding_aes_key, corp_id)
        self.corp_id = corp_id

    def check_signature(self, signature, timestamp, nonce, echo_str):
        return self._check_signature(
            signature,
            timestamp,
            nonce,
            echo_str,
            PrpCrypto
        )

    def encrypt_message(self, msg, nonce, timestamp=None):
        return self._encrypt_message(
            msg,
            nonce,
            timestamp,
            PrpCrypto
        )

    def decrypt_message(self, msg, signature, timestamp, nonce):
        return self._decrypt_message(
            msg,
            signature,
            timestamp,
            nonce,
            PrpCrypto
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import hashlib
import time
import datetime

from six.moves.urllib.parse import quote
from optionaldict import optionaldict
from wechatpyyyy.utils import to_binary
from wechatpyyyy.client.api.base import BaseWeChatAPI


class WeChatCustomService(BaseWeChatAPI):

    def add_account(self, account, nickname, password):
        """
        添加客服账号
        详情请参考
        http://mp.weixin.qq.com/wiki/1/70a29afed17f56d537c833f89be979c9.html

        :param account: 完整客服账号，格式为：账号前缀@公众号微信号
        :param nickname: 客服昵称，最长6个汉字或12个英文字符
        :param password: 客服账号登录密码
        :return: 返回的 JSON 数据包
        """
        password = to_binary(password)
        password = hashlib.md5(password).hexdigest()
        return self._post(
            'https://api.weixin.qq.com/customservice/kfaccount/add',
            data={
                'kf_account': account,
                'nickname': nickname,
                'password': password
            }
        )

    def update_account(self, account, nickname, password):
        """
        更新客服账号
        详情请参考
        http://mp.weixin.qq.com/wiki/1/70a29afed17f56d537c833f89be979c9.html

        :param account: 完整客服账号，格式为：账号前缀@公众号微信号
        :param nickname: 客服昵称，最长6个汉字或12个英文字符
        :param password: 客服账号登录密码
        :return: 返回的 JSON 数据包
        """
        password = to_binary(password)
        password = hashlib.md5(password).hexdigest()
        return self._post(
            'https://api.weixin.qq.com/customservice/kfaccount/update',
            data={
                'kf_account': account,
                'nickname': nickname,
                'password': password
            }
        )

    def delete_account(self, account):
        """
        删除客服账号
        详情请参考
        http://mp.weixin.qq.com/wiki/1/70a29afed17f56d537c833f89be979c9.html

        :param account: 完整客服账号，格式为：账号前缀@公众号微信号
        :return: 返回的 JSON 数据包
        """
        params_data = [
            'access_token={0}'.format(quote(self.access_token)),
            'kf_account={0}'.format(quote(to_binary(account), safe=b'/@')),
        ]
        params = '&'.join(params_data)
        return self._get(
            'https://api.weixin.qq.com/customservice/kfaccount/del',
            params=params
        )

    def get_accounts(self):
        """
        获取客服账号列表
        详情请参考
        http://mp.weixin.qq.com/wiki/1/70a29afed17f56d537c833f89be979c9.html

        :return: 客服账号列表
        """
        res = self._get(
            'customservice/getkflist',
            result_processor=lambda x: x['kf_list']
        )
        return res

    def upload_headimg(self, account, media_file):
        """
        上传客服账号头像
        详情请参考
        http://mp.weixin.qq.com/wiki/1/70a29afed17f56d537c833f89be979c9.html

        :param account: 完整客服账号
        :param media_file: 要上传的头像文件，一个 File-Object
        :return: 返回的 JSON 数据包
        """
        return self._post(
            'https://api.weixin.qq.com/customservice/kfaccount/uploadheadimg',
            params={
                'kf_account': account
            },
            files={
                'media': media_file
            }
        )

    def get_online_accounts(self):
        """
        获取在线客服接待信息
        详情请参考
        http://mp.weixin.qq.com/wiki/9/6fff6f191ef92c126b043ada035cc935.html

        :return: 客服接待信息列表
        """
        res = self._get(
            'customservice/getonlinekflist',
            result_processor=lambda x: x['kf_online_list']
        )
        return res

    def create_session(self, openid, account, text=None):
        """
        多客服创建会话
        详情请参考
        http://mp.weixin.qq.com/wiki/2/6c20f3e323bdf5986cfcb33cbd3b829a.html

        :param openid: 客户 openid
        :param account: 完整客服账号
        :param text: 附加信息，可选
        :return: 返回的 JSON 数据包
        """
        data = optionaldict(
            openid=openid,
            kf_account=account,
            text=text
        )
        return self._post(
            'https://api.weixin.qq.com/customservice/kfsession/create',
            data=data
        )

    def close_session(self, openid, account, text=None):
        """
        多客服关闭会话
        详情请参考
        http://mp.weixin.qq.com/wiki/2/6c20f3e323bdf5986cfcb33cbd3b829a.html

        :param openid: 客户 openid
        :param account: 完整客服账号
        :param text: 附加信息，可选
        :return: 返回的 JSON 数据包
        """
        data = optionaldict(
            openid=openid,
            kf_account=account,
            text=text
        )
        return self._post(
            'https://api.weixin.qq.com/customservice/kfsession/close',
            data=data
        )

    def get_session(self, openid):
        """
        获取客户的会话状态
        详情请参考
        http://mp.weixin.qq.com/wiki/2/6c20f3e323bdf5986cfcb33cbd3b829a.html

        :param openid: 客户 openid
        :return: 返回的 JSON 数据包
        """
        return self._get(
            'https://api.weixin.qq.com/customservice/kfsession/getsession',
            params={'openid': openid}
        )

    def get_session_list(self, account):
        """
        获取客服的会话列表
        详情请参考
        http://mp.weixin.qq.com/wiki/2/6c20f3e323bdf5986cfcb33cbd3b829a.html

        :param account: 完整客服账号
        :return: 客服的会话列表
        """
        res = self._get(
            'https://api.weixin.qq.com/customservice/kfsession/getsessionlist',
            params={'kf_account': account},
            result_processor=lambda x: x['sessionlist']
        )
        return res

    def get_wait_case(self):
        """
        获取未接入会话列表
        详情请参考
        http://mp.weixin.qq.com/wiki/2/6c20f3e323bdf5986cfcb33cbd3b829a.html

        :return: 返回的 JSON 数据包
        """
        return self._get(
            'https://api.weixin.qq.com/customservice/kfsession/getwaitcase'
        )

    def get_records(self, start_time, end_time, msgid=1,
                    number=10000):
        """
        获取客服聊天记录

        :param start_time: 查询开始时间，UNIX 时间戳
        :param end_time: 查询结束时间，UNIX 时间戳，每次查询不能跨日查询
        :param msgid: 消息id顺序从小到大，从1开始
        :param number: 每次获取条数，最多10000条

        :return: 返回的 JSON 数据包
        """
        if isinstance(start_time, datetime.datetime):
            start_time = time.mktime(start_time.timetuple())
        if isinstance(end_time, datetime.datetime):
            end_time = time.mktime(end_time.timetuple())
        record_data = {
            'starttime': int(start_time),
            'endtime': int(end_time),
            'msgid': msgid,
            'number': number
        }
        res = self._post(
            'https://api.weixin.qq.com/customservice/msgrecord/getmsglist',
            data=record_data,
        )
        return res



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from optionaldict import optionaldict

from wechatpyyyy.client.api.base import BaseWeChatAPI


class WeChatWxa(BaseWeChatAPI):
    API_BASE_URL = 'https://api.weixin.qq.com/'

    def create_qrcode(self, path, width=430):
        """
        创建小程序二维码（接口C：适用于需要的码数量较少的业务场景）
        详情请参考
        https://mp.weixin.qq.com/debug/wxadoc/dev/api/qrcode.html
        """
        return self._post(
            'cgi-bin/wxaapp/createwxaqrcode',
            data={
                'path': path,
                'width': width
            }
        )

    def get_wxa_code(self,
                     path,
                     width=430,
                     auto_color=False,
                     line_color={"r": "0", "g": "0", "b": "0"},
                     is_hyaline=False):
        """
        创建小程序码（接口A: 适用于需要的码数量较少的业务场景）
        详情请参考
        https://mp.weixin.qq.com/debug/wxadoc/dev/api/qrcode.html
        """
        return self._post(
            'wxa/getwxacode',
            data={
                'path': path,
                'width': width,
                'auto_color': auto_color,
                'line_color': line_color,
                'is_hyaline': is_hyaline,
            }
        )

    def get_wxa_code_unlimited(self,
                               scene,
                               width=430,
                               auto_color=False,
                               line_color={"r": "0", "g": "0", "b": "0"},
                               page=None,
                               is_hyaline=False):
        """
        创建小程序码（接口B：适用于需要的码数量极多，或仅临时使用的业务场景）
        详情请参考
        https://mp.weixin.qq.com/debug/wxadoc/dev/api/qrcode.html
        """
        return self._post(
            'wxa/getwxacodeunlimit',
            data=optionaldict(
                scene=scene,
                page=page,
                width=width,
                auto_color=auto_color,
                line_color=line_color,
                is_hyaline=is_hyaline,
            )
        )

    def send_template_message(self, user_id, template_id, data, form_id, page=None, color=None, emphasis_keyword=None):
        """
        发送模板消息
        详情请参考
        https://mp.weixin.qq.com/debug/wxadoc/dev/api/notice.html
        """
        tpl_data = optionaldict(
            touser=user_id,
            template_id=template_id,
            page=page,
            form_id=form_id,
            data=data,
            color=color,
            emphasis_keyword=emphasis_keyword,
        )
        return self._post(
            'cgi-bin/message/wxopen/template/send',
            data=tpl_data
        )

    def send_subscribe_message(self, user_id, template_id, data, page=None):
        """
        发送小程序订阅消息
        详情请参考
        https://developers.weixin.qq.com/miniprogram/dev/api-backend/open-api/subscribe-message/subscribeMessage.send.html
        """
        subs_data = optionaldict(
            touser=user_id,
            template_id=template_id,
            page=page,
            data=data,
        )
        return self._post(
            'cgi-bin/message/subscribe/send',
            data=subs_data
        )

    def modify_domain(self, action, request_domain=(), wsrequest_domain=(), upload_domain=(), download_domain=()):
        """
        修改小程序服务器授权域名
        详情请参考
        https://open.weixin.qq.com/cgi-bin/showdocument?action=dir_list&id=open1489138143_WPbOO

        :param action: 增删改查的操作类型，仅支持 'add', 'delete', 'set', 'get'
        :param request_domain: request 合法域名
        :param wsrequest_domain: socket 合法域名
        :param upload_domain: upload file 合法域名
        :param download_domain: download file 合法域名
        """
        return self._post(
            'wxa/modify_domain',
            data={
                'action': action,
                'requestdomain': request_domain,
                'wsrequestdomain': wsrequest_domain,
                'uploaddomain': upload_domain,
                'downloaddomain': download_domain,
            }
        )

    def bind_tester(self, wechat_id):
        """
        绑定微信用户成为小程序体验者
        详情请参考
        https://open.weixin.qq.com/cgi-bin/showdocument?action=dir_list&id=open1489140588_nVUgx

        :param wechat_id: 微信号
        """
        return self._post(
            'wxa/bind_tester',
            data={
                'wechatid': wechat_id,
            }
        )

    def unbind_tester(self, wechat_id):
        """
        解除绑定小程序的体验者
        详情请参考
        https://open.weixin.qq.com/cgi-bin/showdocument?action=dir_list&id=open1489140588_nVUgx

        :param wechat_id: 微信号
        """
        return self._post(
            'wxa/unbind_tester',
            data={
                'wechatid': wechat_id,
            }
        )

    def commit(self, template_id, ext_json, version, description):
        """
        为授权的小程序账号上传小程序代码
        详情请参考
        https://open.weixin.qq.com/cgi-bin/showdocument?action=dir_list&id=open1489140610_Uavc4

        :param template_id: 代码库中的代码模板 ID
        :param ext_json: 第三方自定义的配置
        :param version: 代码版本号，开发者可自定义
        :param description: 代码描述，开发者可自定义
        """
        return self._post(
            'wxa/commit',
            data={
                'template_id': template_id,
                'ext_json': ext_json,
                'user_version': version,
                'user_desc': description,
            },
        )

    def get_qrcode(self):
        """
        获取体验小程序的体验二维码
        返回 Response 类型，header 中带有 Content-Type 与 Content-disposition 类型
        详情请参考
        https://open.weixin.qq.com/cgi-bin/showdocument?action=dir_list&id=open1489140610_Uavc4

        :rtype: requests.Response
        """
        return self._get('wxa/get_qrcode')

    def get_category(self):
        """
        获取授权小程序账号的可选类目
        详情请参考
        https://open.weixin.qq.com/cgi-bin/showdocument?action=dir_list&id=open1489140610_Uavc4

        :rtype: list[dict]
        """
        return self._get(
            'wxa/get_category',
            result_processor=lambda x: x['category_list'],
        )

    def get_page(self):
        """
        获取小程序的第三方提交代码的页面配置
        详情请参考
        https://open.weixin.qq.com/cgi-bin/showdocument?action=dir_list&id=open1489140610_Uavc4

        :rtype: list
        """
        return self._get(
            'wxa/get_page',
            result_processor=lambda x: x['page_list'],
        )

    def submit_audit(self, data):
        """
        将第三方提交的代码包提交审核
        详情请参考
        https://developers.weixin.qq.com/doc/oplatform/Third-party_Platforms/Mini_Programs/code/submit_audit.html
        """
        return self._post(
            'wxa/submit_audit',
            data=data
        )

    def undo_code_audit(self):
        """
        调用本接口可以撤回当前的代码审核单
        注意： 单个帐号每天审核撤回次数最多不超过 1 次，一个月不超过 10 次。
        https://developers.weixin.qq.com/doc/oplatform/Third-party_Platforms/Mini_Programs/code/undocodeaudit.html
        """
        return self._get(
            'wxa/undocodeaudit'
        )

    def revert_code_release(self):
        """
        调用本接口可以将小程序的线上版本进行回退
        注意：
        1. 如果没有上一个线上版本，将无法回退
        2. 只能向上回退一个版本，即当前版本回退后，不能再调用版本回退接口
        https://developers.weixin.qq.com/doc/oplatform/Third-party_Platforms/Mini_Programs/code/revertcoderelease.html
        """
        return self._get(
            'wxa/revertcoderelease'
        )

    def get_audit_status(self, auditid):
        """
        查询某个指定版本的审核状态
        详情请参考
        https://open.weixin.qq.com/cgi-bin/showdocument?action=dir_list&id=open1489140610_Uavc4

        :param auditid: 审核编号
        :type auditid: int
        :return: 一个包含 status, reason 的 dict。status 0为审核成功，1为审核失败，2为审核中。
        """
        return self._post(
            'wxa/get_auditstatus',
            data={
                'auditid': auditid,
            },
        )

    def get_latest_audit_status(self):
        """
        查询最近一次提交的审核状态
        详情请参考
        https://open.weixin.qq.com/cgi-bin/showdocument?action=dir_list&id=open1489140610_Uavc4

        :return: 一个包含 status, reason, auditid 的 dict。status 0为审核成功，1为审核失败，2为审核中。
        """
        return self._get(
            'wxa/get_latest_auditstatus'
        )

    def release(self):
        """
        发布已通过审核的小程序
        详情请参考
        https://open.weixin.qq.com/cgi-bin/showdocument?action=dir_list&id=open1489140610_Uavc4
        """
        return self._post(
            'wxa/release',
            data={},
        )

    def change_visit_status(self, close=False):
        """
        修改小程序线上代码的可见状态
        详情请参考
        https://open.weixin.qq.com/cgi-bin/showdocument?action=dir_list&id=open1489140610_Uavc4

        :param close: close 为 True 时会关闭小程序线上代码的可见状态。
        :type close: bool
        """
        return self._post(
            'wxa/change_visitstatus',
            data={
                'action': 'close' if close else 'open',
            },
        )

    def list_library_templates(self, offset=0, count=20):
        """
        获取小程序模板库里，所有模板的ID与标题
        详情请参考
        https://open.weixin.qq.com/cgi-bin/showdocument?action=dir_list&id=open1500465446_j4CgR

        :param offset: 用于分页，表示起始量，最小值为0
        :type offset: int
        :param count: 用于分页，表示拉取数量，最大值为20
        :type count: int
        :return: 带有 total_count 与 list 的数据
        :rtype: dict
        """
        return self._post(
            'cgi-bin/wxopen/template/library/list',
            data={
                'offset': offset,
                'count': count,
            },
        )

    def get_library_template(self, template_short_id):
        """
        获取小程序模板库里，某个模板的详细信息
        详情请参考
        https://open.weixin.qq.com/cgi-bin/showdocument?action=dir_list&id=open1500465446_j4CgR

        :param template_short_id: 模板标题ID
        :rtype: dict
        """
        return self._post(
            'cgi-bin/wxopen/template/library/get',
            data={
                'id': template_short_id,
            },
        )

    def list_templates(self, offset=0, count=20):
        """
        获取本账号内所有模板
        详情请参考
        https://open.weixin.qq.com/cgi-bin/showdocument?action=dir_list&id=open1500465446_j4CgR

        :param offset: 用于分页，表示起始量，最小值为0
        :type offset: int
        :param count: 用于分页，表示拉取数量，最大值为20
        :type count: int
        :return: 模板列表
        :rtype: list[dict]
        """
        return self._post(
            'cgi-bin/wxopen/template/list',
            data={
                'offset': offset,
                'count': count,
            },
            result_processor=lambda x: x['list'],
        )

    def add_template(self, template_short_id, keyword_id_list):
        """
        组合模板，并将其添加至账号下的模板列表里
        详情请参考
        https://open.weixin.qq.com/cgi-bin/showdocument?action=dir_list&id=open1500465446_j4CgR

        :param template_short_id: 模板标题ID
        :param keyword_id_list: 按照顺序排列的模板关键词列表，最多10个
        :type keyword_id_list: list[int]
        :return: 模板ID
        """
        return self._post(
            'cgi-bin/wxopen/template/add',
            data={
                'id': template_short_id,
                'keyword_id_list': keyword_id_list,
            },
            result_processor=lambda x: x['template_id'],
        )

    def del_template(self, template_id):
        """
        删除本账号内某个模板
        详情请参考
        https://open.weixin.qq.com/cgi-bin/showdocument?action=dir_list&id=open1500465446_j4CgR

        :param template_id: 模板ID
        """
        return self._post(
            'cgi-bin/wxopen/template/del',
            data={
                'template_id': template_id,
            },
        )

    def create_open(self, appid):
        """
        创建开放平台账号，并绑定公众号/小程序
        详情请参考
        https://open.weixin.qq.com/cgi-bin/showdocument?action=dir_list&id=open1498704199_1bcax

        :param appid: 授权公众号或小程序的 appid
        :return: 开放平台的 appid
        """
        return self._post(
            'cgi-bin/open/create',
            data={
                'appid': appid,
            },
            result_processor=lambda x: x['open_appid'],
        )

    def get_open(self, appid):
        """
        获取公众号/小程序所绑定的开放平台账号
        详情请参考
        https://open.weixin.qq.com/cgi-bin/showdocument?action=dir_list&id=open1498704199_1bcax

        :param appid: 授权公众号或小程序的 appid
        :return: 开放平台的 appid
        """
        return self._post(
            'cgi-bin/open/get',
            data={
                'appid': appid,
            },
            result_processor=lambda x: x['open_appid'],
        )

    def bind_open(self, appid, open_appid):
        """
        将公众号/小程序绑定到开放平台帐号下
        详情请参考
        https://open.weixin.qq.com/cgi-bin/showdocument?action=dir_list&id=open1498704199_1bcax

        :param appid: 授权公众号或小程序的 appid
        :param open_appid: 开放平台帐号 appid
        """
        return self._post(
            'cgi-bin/open/bind',
            data={
                'appid': appid,
                'open_appid': open_appid,
            }
        )

    def unbind_open(self, appid, open_appid):
        """
        将公众号/小程序绑定到开放平台帐号下
        详情请参考
        https://open.weixin.qq.com/cgi-bin/showdocument?action=dir_list&id=open1498704199_1bcax

        :param appid: 授权公众号或小程序的 appid
        :param open_appid: 开放平台帐号 appid
        """
        return self._post(
            'cgi-bin/open/unbind',
            data={
                'appid': appid,
                'open_appid': open_appid,
            }
        )

    def code_to_session(self, js_code):
        """
        登录凭证校验。通过 wx.login() 接口获得临时登录凭证 code 后传到开发者服务器调用此接口完成登录流程。更多使用方法详见 小程序登录
        详情请参考
        https://developers.weixin.qq.com/miniprogram/dev/api/code2Session.html

        :param js_code:
        :return:
        """
        return self._get(
            'sns/jscode2session',
            params={
                'appid': self.appid,
                'secret': self.secret,
                'js_code': js_code,
                'grant_type': 'authorization_code'
            }
        )

    def check_image_security(self, media):
        """
        校验一张图片是否含有违法违规内容。
        详情请参考
        https://developers.weixin.qq.com/miniprogram/dev/api-backend/open-api/sec-check/security.imgSecCheck.html

        :param media: 要检测的图片文件，格式支持PNG、JPEG、JPG、GIF，图片尺寸不超过 750px x 1334px
        :return:
        """
        return self._post(
            'wxa/img_sec_check',
            files={
                'media': media,
            }
        )

    def check_text_security(self, content):
        """
        检查一段文本是否含有违法违规内容。
        详情请参考
        https://developers.weixin.qq.com/miniprogram/dev/api-backend/open-api/sec-check/security.msgSecCheck.html

        :param content: 要检测的文本内容，长度不超过 500KB
        :return:
        """
        return self._post(
            'wxa/msg_sec_check',
            data={
                'content': content,
            }
        )

    def check_media_async(self, media_url, media_type):
        """异步校验图片/音频是否含有违法违规内容。
        详情请参考
        https://developers.weixin.qq.com/miniprogram/dev/api-backend/open-api/sec-check/security.mediaCheckAsync.html

        :param media_url: 要检测的多媒体url
        :param media_type: 1:音频;2:图片
        :return:
        """
        return self._post(
            'wxa/media_check_async',
            data={
                'media_url': media_url,
                'media_type': media_type
            }
        )

    def speed_up_audit(self, auditid):
        """
        加急审核申请
        有加急次数的第三方可以通过该接口，对已经提审的小程序进行加急操作，加急后的小程序预计2-12小时内审完。
        https://developers.weixin.qq.com/doc/oplatform/Third-party_Platforms/Mini_Programs/code/speedup_audit.html
        """
        return self._post(
            'wxa/speedupaudit',
            data={
                "auditid": auditid
            }
        )

    def query_quota(self):
        """
        查询服务商的当月提审限额（quota）和加急次数
        服务商可以调用该接口，查询当月平台分配的提审限额和剩余可提审次数，以及当月分配的审核加急次数和剩余加急次数。（所有旗下小程序共用该额度）
        https://developers.weixin.qq.com/doc/oplatform/Third-party_Platforms/Mini_Programs/code/query_quota.html
        """
        return self._get('wxa/queryquota')

    def get_paid_unionid(self, openid, transaction_id=None, mch_id=None, out_trade_no=None):
        """
        用户支付完成后，获取该用户的 UnionId，无需用户授权。

        详情请参考
        https://developers.weixin.qq.com/miniprogram/dev/api-backend/open-api/user-info/auth.getPaidUnionId.html

        :param openid: 支付用户唯一标识
        :param transaction_id: 微信支付订单号，可选
        :param mch_id: 微信支付分配的商户号，和商户订单号配合使用，可选
        :param out_trade_no: 微信支付商户订单号，和商户号配合使用
        :return: 用户唯一标识 unionid
        """
        return self._get(
            'wxa/getpaidunionid',
            params={
                'openid': openid,
                'transaction_id': transaction_id,
                'mch_id': mch_id,
                'out_trade_no': out_trade_no,
            },
            result_processor=lambda x: x['unionid']
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from wechatpyyyy.enterprise.client import WeChatClient  # NOQA
from wechatpyyyy.enterprise.crypto import WeChatCrypto  # NOQA
from wechatpyyyy.enterprise.parser import parse_message  # NOQA
from wechatpyyyy.enterprise.replies import create_reply  # NOQA
